@@ -4,7 +4,9 @@ pub mod gc;
 pub mod hide;
 pub mod hooks;
 pub mod init;
+pub mod r#move;
 pub mod navigation;
 pub mod restack;
 pub mod smartlog;
 pub mod undo;
+pub mod wrap;
