@@ -1,16 +1,16 @@
 //! Install any hooks, aliases, etc. to set up `git-branchless` in this repo.
 
-use std::fmt::Display;
-use std::io::{stdin, stdout, BufRead, BufReader, Write};
+use std::fmt::Write;
+use std::io::{stdin, stdout, BufRead, BufReader, Write as WriteIo};
 use std::path::PathBuf;
 
-use anyhow::Context;
 use console::style;
-use fn_error_context::context;
-use log::warn;
+use eyre::Context;
+use tracing::{instrument, warn};
 
 use crate::core::config::get_core_hooks_path;
-use crate::util::{get_repo, run_git_silent, wrap_git_error, GitRunInfo, GitVersion};
+use crate::git::{Config, ConfigValue, GitRunInfo, GitVersion, Repo};
+use crate::tui::Effects;
 
 const ALL_HOOKS: &[(&str, &str)] = &[
     (
@@ -19,6 +19,12 @@ const ALL_HOOKS: &[(&str, &str)] = &[
 git branchless hook-post-commit "$@"
 "#,
     ),
+    (
+        "post-merge",
+        r#"
+        git branchless hook-post-merge "$@"
+        "#,
+    ),
     (
         "post-rewrite",
         r#"
@@ -72,9 +78,9 @@ enum Hook {
     MultiHook { path: PathBuf },
 }
 
-#[context("Determining hook path")]
-fn determine_hook_path(repo: &git2::Repository, hook_type: &str) -> anyhow::Result<Hook> {
-    let multi_hooks_path = repo.path().join("hooks_multi");
+#[instrument]
+fn determine_hook_path(repo: &Repo, hook_type: &str) -> eyre::Result<Hook> {
+    let multi_hooks_path = repo.get_path().join("hooks_multi");
     let hook = if multi_hooks_path.exists() {
         let path = multi_hooks_path
             .join(format!("{}.d", hook_type))
@@ -116,8 +122,8 @@ fn update_between_lines(lines: &str, updated_lines: &str) -> String {
     new_lines
 }
 
-#[context("Updating hook contents: {:?}", hook)]
-fn update_hook_contents(hook: &Hook, hook_contents: &str) -> anyhow::Result<()> {
+#[instrument]
+fn update_hook_contents(hook: &Hook, hook_contents: &str) -> eyre::Result<()> {
     let (hook_path, hook_contents) = match hook {
         Hook::RegularHook { path } => match std::fs::read_to_string(path) {
             Ok(lines) => {
@@ -132,7 +138,7 @@ fn update_hook_contents(hook: &Hook, hook_contents: &str) -> anyhow::Result<()>
                 (path, hook_contents)
             }
             Err(other) => {
-                return Err(anyhow::anyhow!(other));
+                return Err(eyre::eyre!(other));
             }
         },
         Hook::MultiHook { path } => (path, format!("{}\n{}", SHEBANG, hook_contents)),
@@ -140,50 +146,58 @@ fn update_hook_contents(hook: &Hook, hook_contents: &str) -> anyhow::Result<()>
 
     let hook_dir = hook_path
         .parent()
-        .ok_or_else(|| anyhow::anyhow!("No parent for dir {:?}", hook_path))?;
+        .ok_or_else(|| eyre::eyre!("No parent for dir {:?}", hook_path))?;
     std::fs::create_dir_all(hook_dir)
-        .with_context(|| format!("Creating hook dir {:?}", hook_path))?;
+        .wrap_err_with(|| format!("Creating hook dir {:?}", hook_path))?;
     std::fs::write(hook_path, hook_contents)
-        .with_context(|| format!("Writing hook contents to {:?}", hook_path))?;
+        .wrap_err_with(|| format!("Writing hook contents to {:?}", hook_path))?;
 
     // Setting hook file as executable only supported on Unix systems.
     #[cfg(unix)]
     {
         use std::os::unix::fs::PermissionsExt;
         let metadata = std::fs::metadata(hook_path)
-            .with_context(|| format!("Reading hook permissions for {:?}", hook_path))?;
+            .wrap_err_with(|| format!("Reading hook permissions for {:?}", hook_path))?;
         let mut permissions = metadata.permissions();
         let mode = permissions.mode();
         // Set execute bits.
         let mode = mode | 0o111;
         permissions.set_mode(mode);
         std::fs::set_permissions(hook_path, permissions)
-            .with_context(|| format!("Marking {:?} as executable", hook_path))?;
+            .wrap_err_with(|| format!("Marking {:?} as executable", hook_path))?;
     }
 
     Ok(())
 }
 
-#[context("Installing hook of type: {:?}", hook_type)]
-fn install_hook(repo: &git2::Repository, hook_type: &str, hook_script: &str) -> anyhow::Result<()> {
+#[instrument]
+fn install_hook(repo: &Repo, hook_type: &str, hook_script: &str) -> eyre::Result<()> {
     let hook = determine_hook_path(repo, hook_type)?;
     update_hook_contents(&hook, hook_script)?;
     Ok(())
 }
 
-#[context("Installing all hooks")]
-fn install_hooks(repo: &git2::Repository) -> anyhow::Result<()> {
+#[instrument]
+fn install_hooks(effects: &Effects, repo: &Repo) -> eyre::Result<()> {
     for (hook_type, hook_script) in ALL_HOOKS {
-        println!("Installing hook: {}", hook_type);
+        writeln!(
+            effects.get_output_stream(),
+            "Installing hook: {}",
+            hook_type
+        )?;
         install_hook(repo, hook_type, hook_script)?;
     }
     Ok(())
 }
 
-#[context("Uninstalling all hooks")]
-fn uninstall_hooks(repo: &git2::Repository) -> anyhow::Result<()> {
+#[instrument]
+fn uninstall_hooks(effects: &Effects, repo: &Repo) -> eyre::Result<()> {
     for (hook_type, _hook_script) in ALL_HOOKS {
-        println!("Uninstalling hook: {}", hook_type);
+        writeln!(
+            effects.get_output_stream(),
+            "Uninstalling hook: {}",
+            hook_type
+        )?;
         install_hook(
             repo,
             hook_type,
@@ -196,19 +210,15 @@ fn uninstall_hooks(repo: &git2::Repository) -> anyhow::Result<()> {
     Ok(())
 }
 
-#[context("Installing alias: git {:?} -> git branchless {:?}", from, to)]
-fn install_alias(config: &mut git2::Config, from: &str, to: &str) -> anyhow::Result<()> {
-    config
-        .set_str(
-            format!("alias.{}", from).as_str(),
-            format!("branchless {}", to).as_str(),
-        )
-        .map_err(wrap_git_error)?;
+#[instrument]
+fn install_alias(config: &mut Config, from: &str, to: &str) -> eyre::Result<()> {
+    config.set(format!("alias.{}", from), format!("branchless {}", to))?;
     Ok(())
 }
 
-fn detect_main_branch_name(repo: &git2::Repository) -> Option<String> {
-    [
+#[instrument]
+fn detect_main_branch_name(repo: &Repo) -> eyre::Result<Option<String>> {
+    for branch_name in [
         "master",
         "main",
         "mainline",
@@ -216,42 +226,44 @@ fn detect_main_branch_name(repo: &git2::Repository) -> Option<String> {
         "develop",
         "development",
         "trunk",
-    ]
-    .iter()
-    .find_map(|branch_name| {
+    ] {
         if repo
-            .find_branch(branch_name, git2::BranchType::Local)
-            .is_ok()
+            .find_branch(branch_name, git2::BranchType::Local)?
+            .is_some()
         {
-            Some(branch_name.to_string())
-        } else {
-            None
+            return Ok(Some(branch_name.to_string()));
         }
-    })
+    }
+    Ok(None)
 }
 
-#[context("Installing all aliases")]
+#[instrument]
 fn install_aliases(
-    repo: &mut git2::Repository,
-    config: &mut git2::Config,
+    effects: &Effects,
+    repo: &mut Repo,
+    config: &mut Config,
     git_run_info: &GitRunInfo,
-) -> anyhow::Result<()> {
+) -> eyre::Result<()> {
     for (from, to) in ALL_ALIASES {
-        println!(
+        writeln!(
+            effects.get_output_stream(),
             "Installing alias (non-global): git {} -> git branchless {}",
-            from, to
-        );
+            from,
+            to
+        )?;
         install_alias(config, from, to)?;
     }
 
-    let version_str = run_git_silent(repo, git_run_info, None, &["version"])
-        .with_context(|| "Determining Git version")?;
+    let version_str = git_run_info
+        .run_silent(repo, None, &["version"])
+        .wrap_err("Determining Git version")?;
     let version_str = version_str.trim();
     let version: GitVersion = version_str
         .parse()
-        .with_context(|| format!("Parsing Git version string: {}", version_str))?;
+        .wrap_err_with(|| format!("Parsing Git version string: {}", version_str))?;
     if version < GitVersion(2, 29, 0) {
-        print!(
+        write!(
+            effects.get_output_stream(),
             "\
 {warning_str}: the branchless workflow's `git undo` command requires Git
 v2.29 or later, but your Git version is: {version_str}
@@ -267,132 +279,159 @@ the branchless workflow will work properly.
 ",
             warning_str = style("Warning").yellow().bold(),
             version_str = version_str,
-        );
+        )?;
     }
 
     Ok(())
 }
 
-#[context("Uninstalling all aliases")]
-fn uninstall_aliases(config: &mut git2::Config) -> anyhow::Result<()> {
+#[instrument]
+fn uninstall_aliases(effects: &Effects, config: &mut Config) -> eyre::Result<()> {
     for (from, _to) in ALL_ALIASES {
-        println!("Uninstalling alias (non-global): git {}", from);
+        writeln!(
+            effects.get_output_stream(),
+            "Uninstalling alias (non-global): git {}",
+            from
+        )?;
         config
             .remove(&format!("alias.{}", from))
-            .with_context(|| format!("Uninstalling alias {}", from))?;
+            .wrap_err_with(|| format!("Uninstalling alias {}", from))?;
     }
     Ok(())
 }
 
-#[derive(Debug)]
-enum ConfigValue {
-    Bool(bool),
-    String(String),
-}
-
-impl Display for ConfigValue {
-    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        match self {
-            ConfigValue::Bool(value) => write!(f, "{}", value),
-            ConfigValue::String(value) => write!(f, "{}", value),
-        }
+#[instrument(skip(value))]
+fn set_config(
+    effects: &Effects,
+    config: &mut Config,
+    name: &str,
+    value: impl Into<ConfigValue>,
+) -> eyre::Result<()> {
+    fn inner(
+        effects: &Effects,
+        config: &mut Config,
+        name: &str,
+        value: ConfigValue,
+    ) -> eyre::Result<()> {
+        writeln!(
+            effects.get_output_stream(),
+            "Setting config (non-global): {} = {}",
+            name,
+            value
+        )?;
+        config.set(name, value)?;
+        Ok(())
     }
+
+    inner(effects, config, name, value.into())
 }
 
-#[context("Setting config {}", name)]
-fn set_config(config: &mut git2::Config, name: &str, value: ConfigValue) -> anyhow::Result<()> {
-    println!("Setting config (non-global): {} = {}", name, value);
-    match value {
-        ConfigValue::Bool(value) => config.set_bool(name, value)?,
-        ConfigValue::String(value) => config.set_str(name, &value)?,
-    }
-    Ok(())
-}
-
-#[context("Setting all configs")]
+#[instrument(skip(r#in))]
 fn set_configs(
     r#in: &mut impl BufRead,
-    repo: &git2::Repository,
-    config: &mut git2::Config,
-) -> anyhow::Result<()> {
-    let main_branch_name = match detect_main_branch_name(repo) {
+    effects: &Effects,
+    repo: &Repo,
+    config: &mut Config,
+) -> eyre::Result<()> {
+    let main_branch_name = match detect_main_branch_name(repo)? {
         Some(main_branch_name) => {
-            println!(
+            writeln!(
+                effects.get_output_stream(),
                 "Auto-detected your main branch as: {}",
                 console::style(&main_branch_name).bold()
-            );
-            println!("If this is incorrect, run: git config branchless.core.mainBranch <branch>");
+            )?;
+            writeln!(
+                effects.get_output_stream(),
+                "If this is incorrect, run: git config branchless.core.mainBranch <branch>"
+            )?;
             main_branch_name
         }
         None => {
-            println!(
+            writeln!(
+                effects.get_output_stream(),
                 "{}",
                 console::style("Your main branch name could not be auto-detected!")
                     .yellow()
                     .bold()
-            );
-            println!("Examples of a main branch: master, main, trunk, etc.");
-            println!("See https://github.com/arxanas/git-branchless/wiki/Concepts#main-branch");
-            print!("Enter the name of your main branch: ");
+            )?;
+            writeln!(
+                effects.get_output_stream(),
+                "Examples of a main branch: master, main, trunk, etc."
+            )?;
+            writeln!(
+                effects.get_output_stream(),
+                "See https://github.com/arxanas/git-branchless/wiki/Concepts#main-branch"
+            )?;
+            write!(
+                effects.get_output_stream(),
+                "Enter the name of your main branch: "
+            )?;
             stdout().flush()?;
             let mut input = String::new();
             r#in.read_line(&mut input)?;
             match input.trim() {
-                "" => anyhow::bail!("No main branch name provided"),
+                "" => eyre::bail!("No main branch name provided"),
                 main_branch_name => main_branch_name.to_string(),
             }
         }
     };
     set_config(
+        effects,
         config,
         "branchless.core.mainBranch",
-        ConfigValue::String(main_branch_name),
+        main_branch_name,
     )?;
-    set_config(config, "advice.detachedHead", ConfigValue::Bool(false))?;
+    set_config(effects, config, "advice.detachedHead", false)?;
     Ok(())
 }
 
-#[context("Unsetting all configs")]
-fn unset_configs(config: &mut git2::Config) -> anyhow::Result<()> {
+#[instrument]
+fn unset_configs(effects: &Effects, config: &mut Config) -> eyre::Result<()> {
     for key in ["branchless.core.mainBranch", "advice.detachedHead"] {
-        println!("Unsetting config (non-global): {}", key);
+        writeln!(
+            effects.get_output_stream(),
+            "Unsetting config (non-global): {}",
+            key
+        )?;
         config
             .remove(key)
-            .with_context(|| format!("Unsetting config {}", key))?;
+            .wrap_err_with(|| format!("Unsetting config {}", key))?;
     }
     Ok(())
 }
 
 /// Initialize `git-branchless` in the current repo.
-#[context("Initializing git-branchless for repo")]
-pub fn init(git_run_info: &GitRunInfo) -> anyhow::Result<()> {
+#[instrument]
+pub fn init(effects: &Effects, git_run_info: &GitRunInfo) -> eyre::Result<()> {
     let mut in_ = BufReader::new(stdin());
-    let mut repo = get_repo()?;
-    let mut config = repo.config().with_context(|| "Getting repo config")?;
-    set_configs(&mut in_, &repo, &mut config)?;
-    install_hooks(&repo)?;
-    install_aliases(&mut repo, &mut config, git_run_info)?;
-    println!(
+    let mut repo = Repo::from_current_dir()?;
+    let mut config = repo.get_config()?;
+    set_configs(&mut in_, effects, &repo, &mut config)?;
+    install_hooks(effects, &repo)?;
+    install_aliases(effects, &mut repo, &mut config, git_run_info)?;
+    writeln!(
+        effects.get_output_stream(),
         "{}",
         console::style("Successfully installed git-branchless.")
             .green()
             .bold()
-    );
-    println!(
+    )?;
+    writeln!(
+        effects.get_output_stream(),
         "To uninstall, run: {}",
         console::style("git branchless init --uninstall").bold()
-    );
+    )?;
     Ok(())
 }
 
 /// Uninstall `git-branchless` in the current repo.
-#[context("Uninstall git-branchless for repo")]
-pub fn uninstall() -> anyhow::Result<()> {
-    let repo = get_repo()?;
-    let mut config = repo.config().with_context(|| "Getting repo config")?;
-    unset_configs(&mut config)?;
-    uninstall_hooks(&repo)?;
-    uninstall_aliases(&mut config)?;
+#[instrument]
+pub fn uninstall(effects: &Effects) -> eyre::Result<()> {
+    let repo = Repo::from_current_dir()?;
+    let mut config = repo.get_config().wrap_err("Getting repo config")?;
+    unset_configs(effects, &mut config)?;
+    uninstall_hooks(effects, &repo)?;
+    uninstall_aliases(effects, &mut config)?;
     Ok(())
 }
 
