@@ -4,25 +4,25 @@
 //! log; see the `eventlog` module.
 
 use std::cmp::Ordering;
+use std::fmt::Write;
 use std::time::SystemTime;
 
 use cursive::theme::Effect;
 use cursive::utils::markup::StyledString;
-use fn_error_context::context;
+use tracing::instrument;
 
 use crate::core::eventlog::{EventLogDb, EventReplayer};
 use crate::core::formatting::set_effect;
 use crate::core::formatting::{printable_styled_string, Glyphs, StyledStringBuilder};
 use crate::core::graph::{make_graph, BranchOids, CommitGraph, HeadOid, MainBranchOid};
-use crate::core::mergebase::MergeBaseDb;
+use crate::core::mergebase::{make_merge_base_db, MergeBaseDb};
 use crate::core::metadata::{
     render_commit_metadata, BranchesProvider, CommitMessageProvider, CommitMetadataProvider,
     CommitOidProvider, DifferentialRevisionProvider, HiddenExplanationProvider,
     RelativeTimeProvider,
 };
-use crate::util::{
-    get_branch_oid_to_names, get_db_conn, get_head_oid, get_main_branch_oid, get_repo,
-};
+use crate::git::{NonZeroOid, Repo};
+use crate::tui::Effects;
 
 /// Split fully-independent subgraphs into multiple graphs.
 ///
@@ -32,29 +32,30 @@ use crate::util::{
 /// Returns the list such that the topologically-earlier subgraphs are first in
 /// the list (i.e. those that would be rendered at the bottom of the smartlog).
 fn split_commit_graph_by_roots(
-    repo: &git2::Repository,
-    merge_base_db: &MergeBaseDb,
+    effects: &Effects,
+    repo: &Repo,
+    merge_base_db: &impl MergeBaseDb,
     graph: &CommitGraph,
-) -> Vec<git2::Oid> {
-    let mut root_commit_oids: Vec<git2::Oid> = graph
+) -> Vec<NonZeroOid> {
+    let mut root_commit_oids: Vec<NonZeroOid> = graph
         .iter()
         .filter(|(_oid, node)| node.parent.is_none())
         .map(|(oid, _node)| oid)
         .copied()
         .collect();
 
-    let compare = |lhs_oid: &git2::Oid, rhs_oid: &git2::Oid| -> Ordering {
+    let compare = |lhs_oid: &NonZeroOid, rhs_oid: &NonZeroOid| -> Ordering {
         let lhs_commit = repo.find_commit(*lhs_oid);
         let rhs_commit = repo.find_commit(*rhs_oid);
 
         let (lhs_commit, rhs_commit) = match (lhs_commit, rhs_commit) {
-            (Err(_), Err(_)) | (Err(_), Ok(_)) | (Ok(_), Err(_)) => return lhs_oid.cmp(&rhs_oid),
-            (Ok(lhs_commit), Ok(rhs_commit)) => (lhs_commit, rhs_commit),
+            (Ok(Some(lhs_commit)), Ok(Some(rhs_commit))) => (lhs_commit, rhs_commit),
+            _ => return lhs_oid.cmp(rhs_oid),
         };
 
-        let merge_base_oid = merge_base_db.get_merge_base_oid(repo, *lhs_oid, *rhs_oid);
+        let merge_base_oid = merge_base_db.get_merge_base_oid(effects, repo, *lhs_oid, *rhs_oid);
         let merge_base_oid = match merge_base_oid {
-            Err(_) => return lhs_oid.cmp(&rhs_oid),
+            Err(_) => return lhs_oid.cmp(rhs_oid),
             Ok(merge_base_oid) => merge_base_oid,
         };
 
@@ -66,9 +67,9 @@ fn split_commit_graph_by_roots(
             // The commits were not orderable (pathlogical situation). Let's
             // just order them by timestamp in that case to produce a consistent
             // and reasonable guess at the intended topological ordering.
-            Some(_) | None => match lhs_commit.time().cmp(&rhs_commit.time()) {
+            Some(_) | None => match lhs_commit.get_time().cmp(&rhs_commit.get_time()) {
                 result @ Ordering::Less | result @ Ordering::Greater => result,
-                Ordering::Equal => lhs_oid.cmp(&rhs_oid),
+                Ordering::Equal => lhs_oid.cmp(rhs_oid),
             },
         }
     };
@@ -77,20 +78,20 @@ fn split_commit_graph_by_roots(
     root_commit_oids
 }
 
-#[context("Getting child smartlog output for OID {:?}", &current_oid)]
+#[instrument(skip(commit_metadata_providers, graph))]
 fn get_child_output(
     glyphs: &Glyphs,
     graph: &CommitGraph,
-    root_oids: &[git2::Oid],
+    root_oids: &[NonZeroOid],
     commit_metadata_providers: &mut [&mut dyn CommitMetadataProvider],
     head_oid: &HeadOid,
-    current_oid: git2::Oid,
+    current_oid: NonZeroOid,
     last_child_line_char: Option<&str>,
-) -> anyhow::Result<Vec<StyledString>> {
+) -> eyre::Result<Vec<StyledString>> {
     let current_node = &graph[&current_oid];
     let is_head = {
         let HeadOid(head_oid) = head_oid;
-        Some(current_node.commit.id()) == *head_oid
+        Some(current_node.commit.get_oid()) == *head_oid
     };
 
     let text = render_commit_metadata(&current_node.commit, commit_metadata_providers)?;
@@ -177,18 +178,14 @@ fn get_child_output(
 }
 
 /// Render a pretty graph starting from the given root OIDs in the given graph.
-#[context(
-    "Getting smartlog output for HEAD OID {:?}, root OIDs: {:?}",
-    head_oid,
-    root_oids
-)]
+#[instrument(skip(commit_metadata_providers, graph))]
 fn get_output(
     glyphs: &Glyphs,
     graph: &CommitGraph,
     commit_metadata_providers: &mut [&mut dyn CommitMetadataProvider],
     head_oid: &HeadOid,
-    root_oids: &[git2::Oid],
-) -> anyhow::Result<Vec<StyledString>> {
+    root_oids: &[NonZeroOid],
+) -> eyre::Result<Vec<StyledString>> {
     let mut lines = Vec::new();
 
     // Determine if the provided OID has the provided parent OID as a parent.
@@ -196,16 +193,17 @@ fn get_output(
     // This returns `True` in strictly more cases than checking `graph`,
     // since there may be links between adjacent main branch commits which
     // are not reflected in `graph`.
-    let has_real_parent = |oid: git2::Oid, parent_oid: git2::Oid| -> bool {
+    let has_real_parent = |oid: NonZeroOid, parent_oid: NonZeroOid| -> bool {
         graph[&oid]
             .commit
-            .parent_ids()
+            .get_parent_oids()
+            .into_iter()
             .any(|parent_oid2| parent_oid2 == parent_oid)
     };
 
     for (root_idx, root_oid) in root_oids.iter().enumerate() {
         let root_node = &graph[root_oid];
-        if root_node.commit.parent_count() > 0 {
+        if root_node.commit.get_parent_count() > 0 {
             let line = if root_idx > 0 && has_real_parent(*root_oid, root_oids[root_idx - 1]) {
                 StyledString::plain(glyphs.line.to_owned())
             } else {
@@ -247,17 +245,18 @@ fn get_output(
 }
 
 /// Render the smartlog graph and write it to the provided stream.
+#[instrument(skip(commit_metadata_providers, graph))]
 pub fn render_graph(
-    glyphs: &Glyphs,
-    repo: &git2::Repository,
-    merge_base_db: &MergeBaseDb,
+    effects: &Effects,
+    repo: &Repo,
+    merge_base_db: &impl MergeBaseDb,
     graph: &CommitGraph,
     head_oid: &HeadOid,
     commit_metadata_providers: &mut [&mut dyn CommitMetadataProvider],
-) -> anyhow::Result<Vec<StyledString>> {
-    let root_oids = split_commit_graph_by_roots(repo, merge_base_db, graph);
+) -> eyre::Result<Vec<StyledString>> {
+    let root_oids = split_commit_graph_by_roots(effects, repo, merge_base_db, graph);
     let lines = get_output(
-        glyphs,
+        effects.get_glyphs(),
         graph,
         commit_metadata_providers,
         head_oid,
@@ -267,17 +266,18 @@ pub fn render_graph(
 }
 
 /// Display a nice graph of commits you've recently worked on.
-pub fn smartlog() -> anyhow::Result<()> {
-    let glyphs = Glyphs::detect();
-    let repo = get_repo()?;
-    let conn = get_db_conn(&repo)?;
-    let merge_base_db = MergeBaseDb::new(&conn)?;
+#[instrument]
+pub fn smartlog(effects: &Effects) -> eyre::Result<()> {
+    let repo = Repo::from_current_dir()?;
+    let conn = repo.get_db_conn()?;
     let event_log_db = EventLogDb::new(&conn)?;
-    let event_replayer = EventReplayer::from_event_log_db(&event_log_db)?;
-    let head_oid = get_head_oid(&repo)?;
-    let main_branch_oid = get_main_branch_oid(&repo)?;
-    let branch_oid_to_names = get_branch_oid_to_names(&repo)?;
+    let event_replayer = EventReplayer::from_event_log_db(effects, &repo, &event_log_db)?;
+    let merge_base_db = make_merge_base_db(effects, &repo, &conn, &event_replayer)?;
+    let head_oid = repo.get_head_info()?.oid;
+    let main_branch_oid = repo.get_main_branch_oid()?;
+    let branch_oid_to_names = repo.get_branch_oid_to_names()?;
     let graph = make_graph(
+        effects,
         &repo,
         &merge_base_db,
         &event_replayer,
@@ -289,7 +289,7 @@ pub fn smartlog() -> anyhow::Result<()> {
     )?;
 
     let lines = render_graph(
-        &glyphs,
+        effects,
         &repo,
         &merge_base_db,
         &graph,
@@ -308,7 +308,11 @@ pub fn smartlog() -> anyhow::Result<()> {
         ],
     )?;
     for line in lines {
-        println!("{}", printable_styled_string(&glyphs, line)?);
+        writeln!(
+            effects.get_output_stream(),
+            "{}",
+            printable_styled_string(effects.get_glyphs(), line)?
+        )?;
     }
 
     Ok(())
