@@ -4,40 +4,45 @@
 //! time and inverting them.
 
 use std::convert::TryInto;
-use std::io::{stdin, stdout, BufRead, BufReader, Read, Write};
+use std::ffi::{OsStr, OsString};
+use std::fmt::Write;
+use std::io::{stdin, BufRead, BufReader, Read};
 use std::sync::mpsc::{channel, Receiver, Sender, TryRecvError};
 use std::time::SystemTime;
 
-use anyhow::Context;
 use cursive::event::Key;
+use cursive::traits::Boxable;
 use cursive::utils::markup::StyledString;
-use cursive::views::{Dialog, EditView, LinearLayout, OnEventView, ScrollView, TextView};
+use cursive::views::{Dialog, EditView, LinearLayout, OnEventView, Panel, ScrollView, TextView};
 use cursive::{Cursive, CursiveRunnable, CursiveRunner};
+use eyre::Context;
+use tracing::instrument;
 
 use crate::commands::smartlog::render_graph;
 use crate::core::eventlog::{Event, EventCursor, EventLogDb, EventReplayer, EventTransactionId};
-use crate::core::formatting::{printable_styled_string, Glyphs, Pluralize, StyledStringBuilder};
+use crate::core::formatting::{printable_styled_string, Pluralize, StyledStringBuilder};
 use crate::core::graph::{make_graph, BranchOids, HeadOid, MainBranchOid};
-use crate::core::mergebase::MergeBaseDb;
+use crate::core::mergebase::{make_merge_base_db, MergeBaseDb};
 use crate::core::metadata::{
-    render_commit_metadata, BranchesProvider, CommitMessageProvider, CommitOidProvider,
-    DifferentialRevisionProvider, HiddenExplanationProvider, RelativeTimeProvider,
+    BranchesProvider, CommitMessageProvider, CommitOidProvider, DifferentialRevisionProvider,
+    HiddenExplanationProvider, RelativeTimeProvider,
 };
-use crate::core::tui::{with_siv, SingletonView};
 use crate::declare_views;
-use crate::util::{get_db_conn, get_repo, run_git, GitRunInfo};
+use crate::git::{CategorizedReferenceName, GitRunInfo, MaybeZeroOid, Repo};
+use crate::tui::{with_siv, Effects, SingletonView};
 
 fn render_cursor_smartlog(
-    glyphs: &Glyphs,
-    repo: &git2::Repository,
-    merge_base_db: &MergeBaseDb,
+    effects: &Effects,
+    repo: &Repo,
+    merge_base_db: &impl MergeBaseDb,
     event_replayer: &EventReplayer,
     event_cursor: EventCursor,
-) -> anyhow::Result<Vec<StyledString>> {
+) -> eyre::Result<Vec<StyledString>> {
     let head_oid = event_replayer.get_cursor_head_oid(event_cursor);
     let main_branch_oid = event_replayer.get_cursor_main_branch_oid(event_cursor, repo)?;
     let branch_oid_to_names = event_replayer.get_cursor_branch_oid_to_names(event_cursor, repo)?;
     let graph = make_graph(
+        effects,
         repo,
         merge_base_db,
         event_replayer,
@@ -48,46 +53,28 @@ fn render_cursor_smartlog(
         true,
     )?;
     let result = render_graph(
-        glyphs,
+        effects,
         repo,
         merge_base_db,
         &graph,
         &HeadOid(head_oid),
         &mut [
             &mut CommitOidProvider::new(true)?,
-            &mut RelativeTimeProvider::new(&repo, SystemTime::now())?,
-            &mut HiddenExplanationProvider::new(&graph, &event_replayer, event_cursor)?,
-            &mut BranchesProvider::new(&repo, &branch_oid_to_names)?,
-            &mut DifferentialRevisionProvider::new(&repo)?,
+            &mut RelativeTimeProvider::new(repo, SystemTime::now())?,
+            &mut HiddenExplanationProvider::new(&graph, event_replayer, event_cursor)?,
+            &mut BranchesProvider::new(repo, &branch_oid_to_names)?,
+            &mut DifferentialRevisionProvider::new(repo)?,
             &mut CommitMessageProvider::new()?,
         ],
     )?;
     Ok(result)
 }
 
-fn render_ref_name(ref_name: &str) -> String {
-    match ref_name.strip_prefix("refs/heads/") {
-        Some(branch_name) => format!("branch {}", branch_name),
-        None => format!("ref {}", ref_name),
-    }
-}
+fn describe_event(repo: &Repo, event: &Event) -> eyre::Result<Vec<StyledString>> {
+    // Links to https://github.com/arxanas/git-branchless/issues/57
+    const EMPTY_EVENT_MESSAGE: &str =
+        "This may be an unsupported use-case; see https://git.io/J0b7z";
 
-fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<StyledString>> {
-    let render_commit = |oid: git2::Oid| -> anyhow::Result<StyledString> {
-        match repo.find_commit(oid) {
-            Ok(commit) => render_commit_metadata(
-                &commit,
-                &mut [
-                    &mut CommitOidProvider::new(true)?,
-                    &mut CommitMessageProvider::new()?,
-                ],
-            ),
-            Err(_) => Ok(StyledString::plain(format!(
-                "<unavailable: {} (possibly GC'ed)>",
-                oid.to_string()
-            ))),
-        }
-    };
     let result = match event {
         Event::CommitEvent {
             timestamp: _,
@@ -97,7 +84,7 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Commit ")
-                    .append(render_commit(*commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(*commit_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -107,11 +94,17 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             commit_oid,
+        }
+        | Event::RewriteEvent {
+            timestamp: _,
+            event_tx_id: _,
+            old_commit_oid: MaybeZeroOid::NonZero(commit_oid),
+            new_commit_oid: MaybeZeroOid::Zero,
         } => {
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Hide commit ")
-                    .append(render_commit(*commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(*commit_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -121,11 +114,17 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             commit_oid,
+        }
+        | Event::RewriteEvent {
+            timestamp: _,
+            event_tx_id: _,
+            old_commit_oid: MaybeZeroOid::Zero,
+            new_commit_oid: MaybeZeroOid::NonZero(commit_oid),
         } => {
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Unhide commit ")
-                    .append(render_commit(*commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(*commit_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -135,15 +134,15 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             ref_name,
-            old_ref: None,
-            new_ref: Some(new_ref),
+            old_oid: MaybeZeroOid::Zero,
+            new_oid: MaybeZeroOid::NonZero(new_oid),
             message: _,
         } if ref_name == "HEAD" => {
             // Not sure if this can happen. When a repo is created, maybe?
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Check out to ")
-                    .append(render_commit(new_ref.parse()?)?)
+                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -153,18 +152,18 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             ref_name,
-            old_ref: Some(old_ref),
-            new_ref: Some(new_ref),
+            old_oid: MaybeZeroOid::NonZero(old_oid),
+            new_oid: MaybeZeroOid::NonZero(new_oid),
             message: _,
         } if ref_name == "HEAD" => {
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Check out from ")
-                    .append(render_commit(old_ref.parse()?)?)
+                    .append(repo.friendly_describe_commit_from_oid(*old_oid)?)
                     .build(),
                 StyledStringBuilder::new()
                     .append_plain("            to ")
-                    .append(render_commit(new_ref.parse()?)?)
+                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
                     .build(),
             ]
         }
@@ -173,19 +172,17 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             ref_name,
-            old_ref: None,
-            new_ref: None,
+            old_oid: MaybeZeroOid::Zero,
+            new_oid: MaybeZeroOid::Zero,
             message: _,
         } => {
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Empty event for ")
-                    .append_plain(render_ref_name(ref_name))
+                    .append_plain(CategorizedReferenceName::new(ref_name).render_full())
                     .build(),
                 StyledStringBuilder::new()
-                    .append_plain("This event should not appear. ")
-                    .append_plain("This is a (benign) bug -- ")
-                    .append_plain("please report it.")
+                    .append_plain(EMPTY_EVENT_MESSAGE)
                     .build(),
             ]
         }
@@ -194,16 +191,16 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             ref_name,
-            old_ref: None,
-            new_ref: Some(new_ref),
+            old_oid: MaybeZeroOid::Zero,
+            new_oid: MaybeZeroOid::NonZero(new_oid),
             message: _,
         } => {
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Create ")
-                    .append_plain(render_ref_name(ref_name))
+                    .append_plain(CategorizedReferenceName::new(ref_name).friendly_describe())
                     .append_plain(" at ")
-                    .append(render_commit(new_ref.parse()?)?)
+                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -213,16 +210,16 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             ref_name,
-            old_ref: Some(old_ref),
-            new_ref: None,
+            old_oid: MaybeZeroOid::NonZero(old_oid),
+            new_oid: MaybeZeroOid::Zero,
             message: _,
         } => {
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Delete ")
-                    .append_plain(render_ref_name(ref_name))
+                    .append_plain(CategorizedReferenceName::new(ref_name).friendly_describe())
                     .append_plain(" at ")
-                    .append(render_commit(old_ref.parse()?)?)
+                    .append(repo.friendly_describe_commit_from_oid(*old_oid)?)
                     .build(),
                 StyledString::new(),
             ]
@@ -232,23 +229,23 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
             timestamp: _,
             event_tx_id: _,
             ref_name,
-            old_ref: Some(old_ref),
-            new_ref: Some(new_ref),
+            old_oid: MaybeZeroOid::NonZero(old_oid),
+            new_oid: MaybeZeroOid::NonZero(new_oid),
             message: _,
         } => {
-            let ref_name = render_ref_name(ref_name);
+            let ref_name = CategorizedReferenceName::new(ref_name).friendly_describe();
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Move ")
                     .append_plain(ref_name.clone())
                     .append_plain(" from ")
-                    .append(render_commit(old_ref.parse()?)?)
+                    .append(repo.friendly_describe_commit_from_oid(*old_oid)?)
                     .build(),
                 StyledStringBuilder::new()
                     .append_plain("     ")
                     .append_plain(" ".repeat(ref_name.len()))
                     .append_plain("   to ")
-                    .append(render_commit(new_ref.parse()?)?)
+                    .append(repo.friendly_describe_commit_from_oid(*new_oid)?)
                     .build(),
             ]
         }
@@ -256,17 +253,33 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
         Event::RewriteEvent {
             timestamp: _,
             event_tx_id: _,
-            old_commit_oid,
-            new_commit_oid,
+            old_commit_oid: MaybeZeroOid::NonZero(old_commit_oid),
+            new_commit_oid: MaybeZeroOid::NonZero(new_commit_oid),
         } => {
             vec![
                 StyledStringBuilder::new()
                     .append_plain("Rewrite commit ")
-                    .append(render_commit(*old_commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(*old_commit_oid)?)
                     .build(),
                 StyledStringBuilder::new()
                     .append_plain("           as ")
-                    .append(render_commit(*new_commit_oid)?)
+                    .append(repo.friendly_describe_commit_from_oid(*new_commit_oid)?)
+                    .build(),
+            ]
+        }
+
+        Event::RewriteEvent {
+            timestamp: _,
+            event_tx_id: _,
+            old_commit_oid: MaybeZeroOid::Zero,
+            new_commit_oid: MaybeZeroOid::Zero,
+        } => {
+            vec![
+                StyledStringBuilder::new()
+                    .append_plain("Empty rewrite event.")
+                    .build(),
+                StyledStringBuilder::new()
+                    .append_plain(EMPTY_EVENT_MESSAGE)
                     .build(),
             ]
         }
@@ -275,13 +288,13 @@ fn describe_event(repo: &git2::Repository, event: &Event) -> anyhow::Result<Vec<
 }
 
 fn describe_events_numbered(
-    repo: &git2::Repository,
+    repo: &Repo,
     events: &[Event],
-) -> Result<Vec<StyledString>, anyhow::Error> {
+) -> Result<Vec<StyledString>, eyre::Error> {
     let mut lines = Vec::new();
     for (i, event) in (1..).zip(events) {
         let num_header = format!("{}. ", i);
-        for (j, event_line) in (0..).zip(describe_event(&repo, &event)?) {
+        for (j, event_line) in (0..).zip(describe_event(repo, event)?) {
             let prefix = if j == 0 {
                 num_header.clone()
             } else {
@@ -298,13 +311,14 @@ fn describe_events_numbered(
     Ok(lines)
 }
 
+#[instrument(skip(siv))]
 fn select_past_event(
     mut siv: CursiveRunner<CursiveRunnable>,
-    glyphs: &Glyphs,
-    repo: &git2::Repository,
-    merge_base_db: &MergeBaseDb,
+    effects: &Effects,
+    repo: &Repo,
+    merge_base_db: &impl MergeBaseDb,
     event_replayer: &mut EventReplayer,
-) -> anyhow::Result<Option<EventCursor>> {
+) -> eyre::Result<Option<EventCursor>> {
     #[derive(Clone, Copy, Debug)]
     enum Message {
         Init,
@@ -367,14 +381,9 @@ fn select_past_event(
         let redraw = |siv: &mut Cursive,
                       event_replayer: &mut EventReplayer,
                       event_cursor: EventCursor|
-         -> anyhow::Result<()> {
-            let smartlog = render_cursor_smartlog(
-                &glyphs,
-                &repo,
-                &merge_base_db,
-                &event_replayer,
-                event_cursor,
-            )?;
+         -> eyre::Result<()> {
+            let smartlog =
+                render_cursor_smartlog(effects, repo, merge_base_db, event_replayer, event_cursor)?;
             SmartlogView::find(siv)
                 .get_inner_mut()
                 .set_content(StyledStringBuilder::from_lines(smartlog));
@@ -385,10 +394,7 @@ fn select_past_event(
                     "There are no previous available events.",
                 )],
                 Some((event_id, events)) => {
-                    let event_description = {
-                        let lines = describe_events_numbered(repo, &events)?;
-                        StyledStringBuilder::from_lines(lines)
-                    };
+                    let event_description_lines = describe_events_numbered(repo, events)?;
                     let relative_time_provider = RelativeTimeProvider::new(repo, now)?;
                     let relative_time = if relative_time_provider.is_enabled() {
                         format!(
@@ -401,18 +407,18 @@ fn select_past_event(
                     } else {
                         String::new()
                     };
-                    vec![
-                        StyledStringBuilder::new()
-                            .append_plain("Repo after transaction ")
-                            .append_plain(events[0].get_event_tx_id().to_string())
-                            .append_plain(" (event ")
-                            .append_plain(event_id.to_string())
-                            .append_plain(")")
-                            .append_plain(relative_time)
-                            .append_plain(". Press 'h' for help, 'q' to quit.")
-                            .build(),
-                        event_description,
-                    ]
+
+                    let mut lines = vec![StyledStringBuilder::new()
+                        .append_plain("Repo after transaction ")
+                        .append_plain(events[0].get_event_tx_id().to_string())
+                        .append_plain(" (event ")
+                        .append_plain(event_id.to_string())
+                        .append_plain(")")
+                        .append_plain(relative_time)
+                        .append_plain(". Press 'h' for help, 'q' to quit.")
+                        .build()];
+                    lines.extend(event_description_lines);
+                    lines
                 }
             };
             InfoView::find(siv).set_content(StyledStringBuilder::from_lines(info_view_contents));
@@ -431,10 +437,15 @@ fn select_past_event(
             Ok(Message::Init) => {
                 let smartlog_view: SmartlogView = ScrollView::new(TextView::new("")).into();
                 let info_view: InfoView = TextView::new("").into();
-                siv.add_layer(
+                siv.add_fullscreen_layer(
                     LinearLayout::vertical()
-                        .child(smartlog_view)
-                        .child(info_view),
+                        .child(
+                            Panel::new(smartlog_view)
+                                .title("Commit graph")
+                                .full_height(),
+                        )
+                        .child(Panel::new(ScrollView::new(info_view)).title("Events"))
+                        .full_width(),
                 );
                 redraw(&mut siv, event_replayer, cursor)?;
             }
@@ -524,7 +535,7 @@ fn inverse_event(
     event: Event,
     now: SystemTime,
     event_tx_id: EventTransactionId,
-) -> anyhow::Result<Event> {
+) -> eyre::Result<Event> {
     let timestamp = now.duration_since(SystemTime::UNIX_EPOCH)?.as_secs_f64();
     let inverse_event = match event {
         Event::CommitEvent {
@@ -568,15 +579,15 @@ fn inverse_event(
             timestamp: _,
             event_tx_id: _,
             ref_name,
-            old_ref,
-            new_ref,
+            old_oid: old_ref,
+            new_oid: new_ref,
             message: _,
         } => Event::RefUpdateEvent {
             timestamp,
             event_tx_id,
             ref_name,
-            old_ref: new_ref,
-            new_ref: old_ref,
+            old_oid: new_ref,
+            new_oid: old_ref,
             message: None,
         },
     };
@@ -603,16 +614,16 @@ fn optimize_inverse_events(events: Vec<Event>) -> Vec<Event> {
     optimized_events
 }
 
+#[instrument(skip(in_))]
 fn undo_events(
     in_: &mut impl Read,
-    out: &mut impl Write,
-    glyphs: &Glyphs,
-    repo: &git2::Repository,
+    effects: &Effects,
+    repo: &Repo,
     git_run_info: &GitRunInfo,
     event_log_db: &mut EventLogDb,
     event_replayer: &EventReplayer,
     event_cursor: EventCursor,
-) -> anyhow::Result<isize> {
+) -> eyre::Result<isize> {
     let now = SystemTime::now();
     let event_tx_id = event_log_db.make_transaction_id(now, "undo")?;
     let inverse_events: Vec<Event> = event_replayer
@@ -626,14 +637,14 @@ fn undo_events(
                     timestamp: _,
                     event_tx_id: _,
                     ref_name,
-                    old_ref: None,
-                    new_ref: _,
+                    old_oid: MaybeZeroOid::Zero,
+                    new_oid: _,
                     message: _,
                 } if ref_name == "HEAD"
             )
         })
         .map(|event| inverse_event(event.clone(), now, event_tx_id))
-        .collect::<anyhow::Result<Vec<Event>>>()?;
+        .collect::<eyre::Result<Vec<Event>>>()?;
     let mut inverse_events = optimize_inverse_events(inverse_events);
 
     // Move any checkout operations to be first. Otherwise, we have the risk
@@ -646,19 +657,25 @@ fn undo_events(
     });
 
     if inverse_events.is_empty() {
-        writeln!(out, "No undo actions to apply, exiting.")?;
+        writeln!(
+            effects.get_output_stream(),
+            "No undo actions to apply, exiting."
+        )?;
         return Ok(0);
     }
 
-    writeln!(out, "Will apply these actions:")?;
-    let events = describe_events_numbered(&repo, &inverse_events)?;
+    writeln!(effects.get_output_stream(), "Will apply these actions:")?;
+    let events = describe_events_numbered(repo, &inverse_events)?;
     for line in events {
-        writeln!(out, "{}", printable_styled_string(&glyphs, line)?)?;
+        writeln!(
+            effects.get_output_stream(),
+            "{}",
+            printable_styled_string(effects.get_glyphs(), line)?
+        )?;
     }
 
     let confirmed = {
-        write!(out, "Confirm? [yN] ")?;
-        out.flush()?;
+        write!(effects.get_output_stream(), "Confirm? [yN] ")?;
         let mut user_input = String::new();
         let mut reader = BufReader::new(in_);
         match reader.read_line(&mut user_input) {
@@ -670,7 +687,7 @@ fn undo_events(
         }
     };
     if !confirmed {
-        writeln!(out, "Aborted.")?;
+        writeln!(effects.get_output_stream(), "Aborted.")?;
         return Ok(1);
     }
 
@@ -687,27 +704,33 @@ fn undo_events(
                 timestamp: _,
                 event_tx_id: _,
                 ref_name,
-                old_ref: _,
-                new_ref: Some(new_ref),
+                old_oid: _,
+                new_oid: MaybeZeroOid::NonZero(new_ref),
                 message: _,
             } if ref_name == "HEAD" => {
+                let target_oid: OsString = new_ref.to_string().into();
                 // Most likely the user wanted to perform an actual checkout in
                 // this case, rather than just update `HEAD` (and be left with a
                 // dirty working copy). The `Git` command will update the event
                 // log appropriately, as it will invoke our hooks.
-                run_git(
-                    git_run_info,
-                    Some(event_tx_id),
-                    &["checkout", "--detach", &new_ref],
-                )
-                .with_context(|| "Updating to previous HEAD location")?;
+                git_run_info
+                    .run(
+                        effects,
+                        Some(event_tx_id),
+                        &[
+                            OsStr::new("checkout"),
+                            OsStr::new("--detach"),
+                            target_oid.as_os_str(),
+                        ],
+                    )
+                    .wrap_err("Updating to previous HEAD location")?;
             }
             Event::RefUpdateEvent {
                 timestamp: _,
                 event_tx_id: _,
                 ref_name: _,
-                old_ref: None,
-                new_ref: None,
+                old_oid: MaybeZeroOid::Zero,
+                new_oid: MaybeZeroOid::Zero,
                 message: _,
             } => {
                 // Do nothing.
@@ -716,20 +739,18 @@ fn undo_events(
                 timestamp: _,
                 event_tx_id: _,
                 ref_name,
-                old_ref: Some(_),
-                new_ref: None,
+                old_oid: MaybeZeroOid::NonZero(_),
+                new_oid: MaybeZeroOid::Zero,
                 message: _,
-            } => match repo.find_reference(&ref_name) {
-                Ok(mut reference) => {
-                    reference
-                        .delete()
-                        .with_context(|| format!("Deleting reference: {}", ref_name))?;
+            } => match repo.find_reference(&ref_name)? {
+                Some(mut reference) => {
+                    reference.delete().wrap_err("Applying `RefUpdateEvent`")?;
                 }
-                Err(_) => {
+                None => {
                     writeln!(
-                        out,
+                        effects.get_output_stream(),
                         "Reference {} did not exist, not deleting it.",
-                        ref_name
+                        ref_name.to_string_lossy()
                     )?;
                 }
             },
@@ -737,21 +758,20 @@ fn undo_events(
                 timestamp: _,
                 event_tx_id: _,
                 ref_name,
-                old_ref: None,
-                new_ref: Some(new_ref),
+                old_oid: MaybeZeroOid::Zero,
+                new_oid: MaybeZeroOid::NonZero(new_oid),
                 message: _,
             }
             | Event::RefUpdateEvent {
                 timestamp: _,
                 event_tx_id: _,
                 ref_name,
-                old_ref: Some(_),
-                new_ref: Some(new_ref),
+                old_oid: MaybeZeroOid::NonZero(_),
+                new_oid: MaybeZeroOid::NonZero(new_oid),
                 message: _,
             } => {
                 // Create or update the given reference.
-                let new_ref = new_ref.parse()?;
-                repo.reference(&ref_name, new_ref, true, "branchless undo")?;
+                repo.create_reference(&ref_name, new_oid, true, "branchless undo")?;
             }
             Event::CommitEvent { .. }
             | Event::HideEvent { .. }
@@ -762,22 +782,26 @@ fn undo_events(
         }
     }
 
-    writeln!(out, "Applied {}.", num_inverse_events)?;
+    writeln!(
+        effects.get_output_stream(),
+        "Applied {}.",
+        num_inverse_events
+    )?;
     Ok(0)
 }
 
 /// Restore the repository to a previous state interactively.
-pub fn undo(git_run_info: &GitRunInfo) -> anyhow::Result<isize> {
-    let glyphs = Glyphs::detect();
-    let repo = get_repo()?;
-    let conn = get_db_conn(&repo)?;
-    let merge_base_db = MergeBaseDb::new(&conn)?;
+#[instrument]
+pub fn undo(effects: &Effects, git_run_info: &GitRunInfo) -> eyre::Result<isize> {
+    let repo = Repo::from_current_dir()?;
+    let conn = repo.get_db_conn()?;
     let mut event_log_db = EventLogDb::new(&conn)?;
-    let mut event_replayer = EventReplayer::from_event_log_db(&event_log_db)?;
+    let mut event_replayer = EventReplayer::from_event_log_db(effects, &repo, &event_log_db)?;
+    let merge_base_db = make_merge_base_db(effects, &repo, &conn, &event_replayer)?;
 
     let event_cursor = {
-        let result = with_siv(|siv| {
-            select_past_event(siv, &glyphs, &repo, &merge_base_db, &mut event_replayer)
+        let result = with_siv(effects, |effects, siv| {
+            select_past_event(siv, &effects, &repo, &merge_base_db, &mut event_replayer)
         })?;
         match result {
             Some(event_cursor) => event_cursor,
@@ -787,10 +811,9 @@ pub fn undo(git_run_info: &GitRunInfo) -> anyhow::Result<isize> {
 
     let result = undo_events(
         &mut stdin(),
-        &mut stdout().lock(),
-        &glyphs,
+        effects,
         &repo,
-        &git_run_info,
+        git_run_info,
         &mut event_log_db,
         &event_replayer,
         event_cursor,
@@ -800,39 +823,37 @@ pub fn undo(git_run_info: &GitRunInfo) -> anyhow::Result<isize> {
 
 #[allow(missing_docs)]
 pub mod testing {
-    use std::io::{Read, Write};
+    use std::io::Read;
 
     use cursive::{CursiveRunnable, CursiveRunner};
 
     use crate::core::eventlog::{EventCursor, EventLogDb, EventReplayer};
-    use crate::core::formatting::Glyphs;
     use crate::core::mergebase::MergeBaseDb;
-    use crate::util::GitRunInfo;
+    use crate::git::{GitRunInfo, Repo};
+    use crate::tui::Effects;
 
     pub fn select_past_event(
         siv: CursiveRunner<CursiveRunnable>,
-        glyphs: &Glyphs,
-        repo: &git2::Repository,
-        merge_base_db: &MergeBaseDb,
+        effects: &Effects,
+        repo: &Repo,
+        merge_base_db: &impl MergeBaseDb,
         event_replayer: &mut EventReplayer,
-    ) -> anyhow::Result<Option<EventCursor>> {
-        super::select_past_event(siv, glyphs, repo, merge_base_db, event_replayer)
+    ) -> eyre::Result<Option<EventCursor>> {
+        super::select_past_event(siv, effects, repo, merge_base_db, event_replayer)
     }
 
     pub fn undo_events(
         in_: &mut impl Read,
-        out: &mut impl Write,
-        glyphs: &Glyphs,
-        repo: &git2::Repository,
+        effects: &Effects,
+        repo: &Repo,
         git_run_info: &GitRunInfo,
         event_log_db: &mut EventLogDb,
         event_replayer: &EventReplayer,
         event_cursor: EventCursor,
-    ) -> anyhow::Result<isize> {
+    ) -> eyre::Result<isize> {
         super::undo_events(
             in_,
-            out,
-            glyphs,
+            effects,
             repo,
             git_run_info,
             event_log_db,
@@ -849,32 +870,32 @@ mod tests {
     use crate::core::eventlog::testing::make_dummy_transaction_id;
 
     #[test]
-    fn test_optimize_inverse_events() -> anyhow::Result<()> {
+    fn test_optimize_inverse_events() -> eyre::Result<()> {
         let event_tx_id = make_dummy_transaction_id(123);
         let input = vec![
             Event::RefUpdateEvent {
                 timestamp: 1.0,
                 event_tx_id,
-                ref_name: "HEAD".to_owned(),
-                old_ref: Some("1".parse()?),
-                new_ref: Some("2".parse()?),
+                ref_name: "HEAD".into(),
+                old_oid: MaybeZeroOid::NonZero("1".parse()?),
+                new_oid: MaybeZeroOid::NonZero("2".parse()?),
                 message: None,
             },
             Event::RefUpdateEvent {
                 timestamp: 2.0,
                 event_tx_id,
-                ref_name: "HEAD".to_owned(),
-                old_ref: Some("1".parse()?),
-                new_ref: Some("3".parse()?),
+                ref_name: "HEAD".into(),
+                old_oid: MaybeZeroOid::NonZero("1".parse()?),
+                new_oid: MaybeZeroOid::NonZero("3".parse()?),
                 message: None,
             },
         ];
         let expected = vec![Event::RefUpdateEvent {
             timestamp: 2.0,
             event_tx_id,
-            ref_name: "HEAD".to_owned(),
-            old_ref: Some("1".parse()?),
-            new_ref: Some("3".parse()?),
+            ref_name: "HEAD".into(),
+            old_oid: MaybeZeroOid::NonZero("1".parse()?),
+            new_oid: MaybeZeroOid::NonZero("3".parse()?),
             message: None,
         }];
         assert_eq!(optimize_inverse_events(input), expected);
