@@ -8,29 +8,31 @@
 use std::cmp::Ordering;
 use std::collections::{HashMap, HashSet};
 use std::convert::{TryFrom, TryInto};
+use std::ffi::{OsStr, OsString};
+
 use std::str::FromStr;
 use std::time::{Duration, SystemTime};
 
-use anyhow::Context;
-use fn_error_context::context;
-use log::warn;
+use eyre::Context;
+use tracing::{error, instrument};
 
-use crate::core::config::get_main_branch_name;
-use crate::util::{get_main_branch_oid, wrap_git_error};
+use crate::git::{CategorizedReferenceName, MaybeZeroOid, NonZeroOid, Repo};
+use crate::tui::{Effects, OperationType};
 
 /// When this environment variable is set, we reuse the ID for the transaction
 /// which the caller has already started.
 pub const BRANCHLESS_TRANSACTION_ID_ENV_VAR: &str = "BRANCHLESS_TRANSACTION_ID";
 
 // Wrapper around the row stored directly in the database.
+#[derive(Clone, Debug)]
 struct Row {
     timestamp: f64,
     type_: String,
     event_tx_id: isize,
-    ref1: Option<String>,
-    ref2: Option<String>,
-    ref_name: Option<String>,
-    message: Option<String>,
+    ref1: Option<OsString>,
+    ref2: Option<OsString>,
+    ref_name: Option<OsString>,
+    message: Option<OsString>,
 }
 
 /// The ID associated with the transactions that created an event.
@@ -85,10 +87,10 @@ pub enum Event {
         event_tx_id: EventTransactionId,
 
         /// The OID of the commit before the rewrite.
-        old_commit_oid: git2::Oid,
+        old_commit_oid: MaybeZeroOid,
 
         /// The OID of the commit after the rewrite.
-        new_commit_oid: git2::Oid,
+        new_commit_oid: MaybeZeroOid,
     },
 
     /// Indicates that a reference was updated.
@@ -106,24 +108,16 @@ pub enum Event {
         /// The full name of the reference that was updated.
         ///
         /// For example, `HEAD` or `refs/heads/master`.
-        ref_name: String,
+        ref_name: OsString,
 
-        /// The old referent.
-        ///
-        /// May be an OID (in the case of a direct reference) or another
-        /// reference name (in the case of a symbolic reference).
-        old_ref: Option<String>,
+        /// The old referent OID.
+        old_oid: MaybeZeroOid,
 
-        /// The updated referent.
-        ///
-        /// This may not be different from the old referent.
-        ///
-        /// May be an OID (in the case of a direct reference) or another
-        /// reference name (in the case of a symbolic reference).
-        new_ref: Option<String>,
+        /// The updated referent OID.
+        new_oid: MaybeZeroOid,
 
         /// A message associated with the rewrite, if any.
-        message: Option<String>,
+        message: Option<OsString>,
     },
 
     /// Indicate that the user made a commit.
@@ -137,7 +131,7 @@ pub enum Event {
         event_tx_id: EventTransactionId,
 
         /// The new commit OID.
-        commit_oid: git2::Oid,
+        commit_oid: NonZeroOid,
     },
 
     /// Indicates that a commit was explicitly hidden by the user.
@@ -152,7 +146,7 @@ pub enum Event {
         event_tx_id: EventTransactionId,
 
         /// The OID of the commit that was hidden.
-        commit_oid: git2::Oid,
+        commit_oid: NonZeroOid,
     },
 
     /// Indicates that a commit was explicitly un-hidden by the user.
@@ -167,7 +161,7 @@ pub enum Event {
         event_tx_id: EventTransactionId,
 
         /// The OID of the commit that was unhidden.
-        commit_oid: git2::Oid,
+        commit_oid: NonZeroOid,
     },
 }
 
@@ -208,8 +202,8 @@ impl From<Event> for Row {
                 timestamp,
                 event_tx_id,
                 type_: String::from("rewrite"),
-                ref1: Some(old_commit_oid.to_string()),
-                ref2: Some(new_commit_oid.to_string()),
+                ref1: Some(old_commit_oid.to_string().into()),
+                ref2: Some(new_commit_oid.to_string().into()),
                 ref_name: None,
                 message: None,
             },
@@ -218,15 +212,15 @@ impl From<Event> for Row {
                 timestamp,
                 event_tx_id: EventTransactionId(event_tx_id),
                 ref_name,
-                old_ref,
-                new_ref,
+                old_oid,
+                new_oid,
                 message,
             } => Row {
                 timestamp,
                 event_tx_id,
                 type_: String::from("ref-move"),
-                ref1: old_ref,
-                ref2: new_ref,
+                ref1: Some(old_oid.to_string().into()),
+                ref2: Some(new_oid.to_string().into()),
                 ref_name: Some(ref_name),
                 message,
             },
@@ -239,7 +233,7 @@ impl From<Event> for Row {
                 timestamp,
                 event_tx_id,
                 type_: String::from("commit"),
-                ref1: Some(commit_oid.to_string()),
+                ref1: Some(commit_oid.to_string().into()),
                 ref2: None,
                 ref_name: None,
                 message: None,
@@ -253,7 +247,7 @@ impl From<Event> for Row {
                 timestamp,
                 event_tx_id,
                 type_: String::from("hide"),
-                ref1: Some(commit_oid.to_string()),
+                ref1: Some(commit_oid.to_string().into()),
                 ref2: None,
                 ref_name: None,
                 message: None,
@@ -267,7 +261,7 @@ impl From<Event> for Row {
                 timestamp,
                 event_tx_id,
                 type_: String::from("unhide"),
-                ref1: Some(commit_oid.to_string()),
+                ref1: Some(commit_oid.to_string().into()),
                 ref2: None,
                 ref_name: None,
                 message: None,
@@ -276,91 +270,106 @@ impl From<Event> for Row {
     }
 }
 
-impl TryFrom<Row> for Event {
-    type Error = anyhow::Error;
+fn try_from_row_helper(row: &Row) -> Result<Event, eyre::Error> {
+    let row: Row = row.clone();
+    let Row {
+        timestamp,
+        event_tx_id,
+        type_,
+        ref_name,
+        ref1,
+        ref2,
+        message,
+    } = row;
+    let event_tx_id = EventTransactionId(event_tx_id);
 
-    #[context("Converting database result row into `Event`")]
-    fn try_from(row: Row) -> Result<Self, Self::Error> {
-        let Row {
-            timestamp,
-            event_tx_id,
-            type_,
-            ref_name,
-            ref1,
-            ref2,
-            message,
-        } = row;
-        let event_tx_id = EventTransactionId(event_tx_id);
-
-        let get_oid = |oid: &Option<String>, oid_name: &str| -> anyhow::Result<git2::Oid> {
-            match oid {
-                Some(oid) => {
-                    let oid = git2::Oid::from_str(&oid).map_err(wrap_git_error)?;
+    let get_oid = |oid_str: &Option<OsString>, oid_name: &str| -> eyre::Result<MaybeZeroOid> {
+        match oid_str {
+            Some(oid_str) => match oid_str.to_str() {
+                Some(oid_str) => {
+                    let oid: MaybeZeroOid = oid_str.parse()?;
                     Ok(oid)
                 }
-                None => Err(anyhow::anyhow!(
-                    "OID '{}' was `None` for event type '{}'",
-                    oid_name,
-                    type_
-                )),
-            }
-        };
+                None => eyre::bail!("Invalid OID for reference name: {:?}", oid_str),
+            },
+            None => Err(eyre::eyre!(
+                "OID '{}' was `None` for event type '{}'",
+                oid_name,
+                type_
+            )),
+        }
+    };
 
-        let event = match type_.as_str() {
-            "rewrite" => {
-                let old_commit_oid = get_oid(&ref1, "old commit OID")?;
-                let new_commit_oid = get_oid(&ref2, "new commit OID")?;
-                Event::RewriteEvent {
-                    timestamp,
-                    event_tx_id,
-                    old_commit_oid,
-                    new_commit_oid,
-                }
+    let event = match type_.as_str() {
+        "rewrite" => {
+            let old_commit_oid = get_oid(&ref1, "old commit OID")?;
+            let new_commit_oid = get_oid(&ref2, "new commit OID")?;
+            Event::RewriteEvent {
+                timestamp,
+                event_tx_id,
+                old_commit_oid,
+                new_commit_oid,
             }
+        }
 
-            "ref-move" => {
-                let ref_name =
-                    ref_name.ok_or_else(|| anyhow::anyhow!("ref-move event missing ref name"))?;
-                Event::RefUpdateEvent {
-                    timestamp,
-                    event_tx_id,
-                    ref_name,
-                    old_ref: ref1,
-                    new_ref: ref2,
-                    message,
-                }
+        "ref-move" => {
+            let ref_name =
+                ref_name.ok_or_else(|| eyre::eyre!("ref-move event missing ref name"))?;
+            let ref1 = ref1.ok_or_else(|| eyre::eyre!("ref-move event missing ref1"))?;
+            let ref2 = ref2.ok_or_else(|| eyre::eyre!("ref-move event missing ref2"))?;
+            Event::RefUpdateEvent {
+                timestamp,
+                event_tx_id,
+                ref_name,
+                old_oid: ref1.try_into()?,
+                new_oid: ref2.try_into()?,
+                message,
             }
+        }
 
-            "commit" => {
-                let commit_oid = get_oid(&ref1, "commit OID")?;
-                Event::CommitEvent {
-                    timestamp,
-                    event_tx_id,
-                    commit_oid,
-                }
+        "commit" => {
+            let commit_oid: NonZeroOid = get_oid(&ref1, "commit OID")?.try_into()?;
+            Event::CommitEvent {
+                timestamp,
+                event_tx_id,
+                commit_oid,
             }
+        }
 
-            "hide" => {
-                let commit_oid = get_oid(&ref1, "commit OID")?;
-                Event::HideEvent {
-                    timestamp,
-                    event_tx_id,
-                    commit_oid,
-                }
+        "hide" => {
+            let commit_oid: NonZeroOid = get_oid(&ref1, "commit OID")?.try_into()?;
+            Event::HideEvent {
+                timestamp,
+                event_tx_id,
+                commit_oid,
             }
+        }
 
-            "unhide" => {
-                let commit_oid = get_oid(&ref1, "commit OID")?;
-                Event::UnhideEvent {
-                    timestamp,
-                    event_tx_id,
-                    commit_oid,
-                }
+        "unhide" => {
+            let commit_oid: NonZeroOid = get_oid(&ref1, "commit OID")?.try_into()?;
+            Event::UnhideEvent {
+                timestamp,
+                event_tx_id,
+                commit_oid,
             }
+        }
 
-            other => anyhow::bail!("Unknown event type {}", other),
-        };
-        Ok(event)
+        other => eyre::bail!("Unknown event type {}", other),
+    };
+    Ok(event)
+}
+
+impl TryFrom<Row> for Event {
+    type Error = eyre::Error;
+
+    fn try_from(row: Row) -> Result<Self, Self::Error> {
+        match try_from_row_helper(&row) {
+            Ok(event) => Ok(event),
+            Err(err) => {
+                error!(?row, "Could not convert row into event");
+                Err(err)
+            }
+        }
     }
 }
 
@@ -369,8 +378,14 @@ pub struct EventLogDb<'conn> {
     conn: &'conn rusqlite::Connection,
 }
 
-#[context("Initializing `EventLogDb` tables")]
-fn init_tables(conn: &rusqlite::Connection) -> anyhow::Result<()> {
+impl std::fmt::Debug for EventLogDb<'_> {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(f, "<EventLogDb>")
+    }
+}
+
+#[instrument]
+fn init_tables(conn: &rusqlite::Connection) -> eyre::Result<()> {
     conn.execute(
         "
 CREATE TABLE IF NOT EXISTS event_log (
@@ -385,7 +400,7 @@ CREATE TABLE IF NOT EXISTS event_log (
 ",
         rusqlite::params![],
     )
-    .context("Creating `event_log` table")?;
+    .wrap_err("Creating `event_log` table")?;
 
     conn.execute(
         "
@@ -403,16 +418,16 @@ CREATE TABLE IF NOT EXISTS event_transactions (
 ",
         rusqlite::params![],
     )
-    .context("Creating `event_transactions` table")?;
+    .wrap_err("Creating `event_transactions` table")?;
 
     Ok(())
 }
 
 impl<'conn> EventLogDb<'conn> {
     /// Constructor.
-    #[context("Constructing `EventLogDb`")]
-    pub fn new(conn: &'conn rusqlite::Connection) -> anyhow::Result<Self> {
-        init_tables(&conn)?;
+    #[instrument]
+    pub fn new(conn: &'conn rusqlite::Connection) -> eyre::Result<Self> {
+        init_tables(conn)?;
         Ok(EventLogDb { conn })
     }
 
@@ -420,8 +435,8 @@ impl<'conn> EventLogDb<'conn> {
     ///
     /// Args:
     /// * events: The events to add.
-    #[context("Adding events to event-log")]
-    pub fn add_events(&mut self, events: Vec<Event>) -> anyhow::Result<()> {
+    #[instrument]
+    pub fn add_events(&mut self, events: Vec<Event>) -> eyre::Result<()> {
         let tx = self.conn.unchecked_transaction()?;
         for event in events {
             let Row {
@@ -433,7 +448,16 @@ impl<'conn> EventLogDb<'conn> {
                 ref_name,
                 message,
             } = Row::from(event);
-            tx.execute_named(
+
+            // FIXME: it would be ideal to use BLOBs to store the reference
+            // names instead of TEXT, so that we can represent esoteric
+            // reference names (which are derived from path names).
+            let ref1 = ref1.map(|x| x.to_string_lossy().into_owned());
+            let ref2 = ref2.map(|x| x.to_string_lossy().into_owned());
+            let ref_name = ref_name.map(|x| x.to_string_lossy().into_owned());
+            let message = message.map(|x| x.to_string_lossy().into_owned());
+
+            tx.execute(
                 "
 INSERT INTO event_log VALUES (
     :timestamp,
@@ -463,8 +487,9 @@ INSERT INTO event_log VALUES (
     /// Get all the events in the database.
     ///
     /// Returns: All the events in the database, ordered from oldest to newest.
-    #[context("Querying events from `EventLogDb`")]
-    pub fn get_events(&self) -> anyhow::Result<Vec<Event>> {
+    #[instrument]
+
+    pub fn get_events(&self) -> eyre::Result<Vec<Event>> {
         let mut stmt = self.conn.prepare(
             "
 SELECT timestamp, type, event_tx_id, old_ref, new_ref, ref_name, message
@@ -482,19 +507,14 @@ ORDER BY rowid ASC
                 let new_ref: Option<String> = row.get("new_ref")?;
                 let message: Option<String> = row.get("message")?;
 
-                // A ref name corresponding to commit hash `0` indicates that
-                // there was no old/new ref at all (i.e. it was created or deleted).
-                let old_ref = old_ref.filter(|old_ref| *old_ref != git2::Oid::zero().to_string());
-                let new_ref = new_ref.filter(|new_ref| *new_ref != git2::Oid::zero().to_string());
-
                 Ok(Row {
                     timestamp,
                     event_tx_id,
                     type_,
-                    ref_name,
-                    ref1: old_ref,
-                    ref2: new_ref,
-                    message,
+                    ref_name: ref_name.map(OsString::from),
+                    ref1: old_ref.map(OsString::from),
+                    ref2: new_ref.map(OsString::from),
+                    message: message.map(OsString::from),
                 })
             })?
             .collect();
@@ -502,14 +522,12 @@ ORDER BY rowid ASC
         rows.into_iter().map(Event::try_from).collect()
     }
 
-    /// Create a new event transaction ID to be used to insert subsequent
-    /// `Event`s into the database.
-    #[context("Creating a new `EventTransactionId`")]
-    pub fn make_transaction_id(
+    #[instrument]
+    fn make_transaction_id_inner(
         &self,
         now: SystemTime,
-        message: impl AsRef<str>,
-    ) -> anyhow::Result<EventTransactionId> {
+        message: &str,
+    ) -> eyre::Result<EventTransactionId> {
         if let Ok(transaction_id) = std::env::var(BRANCHLESS_TRANSACTION_ID_ENV_VAR) {
             if let Ok(transaction_id) = transaction_id.parse::<EventTransactionId>() {
                 return Ok(transaction_id);
@@ -520,10 +538,10 @@ ORDER BY rowid ASC
 
         let timestamp = now
             .duration_since(SystemTime::UNIX_EPOCH)
-            .with_context(|| format!("Calculating event transaction timestamp: {:?}", &now))?
+            .wrap_err("Calculating event transaction timestamp")?
             .as_secs_f64();
         self.conn
-            .execute_named(
+            .execute(
                 "
             INSERT INTO event_transactions
             (timestamp, message)
@@ -532,16 +550,10 @@ ORDER BY rowid ASC
         ",
                 rusqlite::named_params! {
                     ":timestamp": timestamp,
-                    ":message": message.as_ref(),
+                    ":message": message,
                 },
             )
-            .with_context(|| {
-                format!(
-                    "Creating event transaction (now: {:?}, message: {:?})",
-                    &now,
-                    message.as_ref(),
-                )
-            })?;
+            .wrap_err("Creating event transaction")?;
 
         // Ensure that we query `last_insert_rowid` in a transaction, in case
         // there's another thread in this process making queries with the same
@@ -550,6 +562,16 @@ ORDER BY rowid ASC
         tx.commit()?;
         Ok(EventTransactionId(event_tx_id))
     }
+
+    /// Create a new event transaction ID to be used to insert subsequent
+    /// `Event`s into the database.
+    pub fn make_transaction_id(
+        &self,
+        now: SystemTime,
+        message: impl AsRef<str>,
+    ) -> eyre::Result<EventTransactionId> {
+        self.make_transaction_id_inner(now, message.as_ref())
+    }
 }
 
 /// Determine whether a given reference is used to keep a commit alive.
@@ -560,8 +582,11 @@ ORDER BY rowid ASC
 /// Returns: Whether or not the given reference is used internally to keep the
 /// commit alive, so that it's not collected by Git's garbage collection
 /// mechanism.
-pub fn is_gc_ref(ref_name: &str) -> bool {
-    ref_name.starts_with("refs/branchless/")
+pub fn is_gc_ref(ref_name: &OsStr) -> bool {
+    match ref_name.to_str() {
+        None => false,
+        Some(ref_name) => ref_name.starts_with("refs/branchless/"),
+    }
 }
 
 /// Determines whether or not updates to the given reference should be ignored.
@@ -570,14 +595,14 @@ pub fn is_gc_ref(ref_name: &str) -> bool {
 /// * `ref_name`: The name of the reference to check.
 ///
 /// Returns: Whether or not updates to the given reference should be ignored.
-pub fn should_ignore_ref_updates(ref_name: &str) -> bool {
+pub fn should_ignore_ref_updates(ref_name: &OsStr) -> bool {
     if is_gc_ref(ref_name) {
         return true;
     }
 
     matches!(
-        ref_name,
-        "ORIG_HEAD" | "CHERRY_PICK" | "REBASE_HEAD" | "CHERRY_PICK_HEAD" | "FETCH_HEAD"
+        ref_name.to_str(),
+        Some("ORIG_HEAD" | "CHERRY_PICK" | "REBASE_HEAD" | "CHERRY_PICK_HEAD" | "FETCH_HEAD")
     )
 }
 
@@ -623,7 +648,6 @@ pub struct EventCursor {
 }
 
 /// Processes events in order and determine the repo's visible commits.
-#[derive(Debug)]
 pub struct EventReplayer {
     /// Events are numbered starting from zero.
     id_counter: isize,
@@ -631,35 +655,61 @@ pub struct EventReplayer {
     /// The list of observed events.
     events: Vec<Event>,
 
+    /// The name of the reference representing the main branch.
+    main_branch_reference_name: OsString,
+
     /// The events that have affected each commit.
-    commit_history: HashMap<git2::Oid, Vec<EventInfo>>,
+    commit_history: HashMap<NonZeroOid, Vec<EventInfo>>,
 
     /// Map from ref names to ref locations (an OID or another ref name). Works
     /// around https://github.com/arxanas/git-branchless/issues/7.
     ///
     /// If an entry is not present, it was either never observed, or it most
     /// recently changed to point to the zero hash (i.e. it was deleted).
-    ref_locations: HashMap<String, String>,
+    ref_locations: HashMap<OsString, NonZeroOid>,
+}
+
+impl std::fmt::Debug for EventReplayer {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(
+            f,
+            "<EventReplayer events.len={:?} ref_locations.len={:?}>",
+            self.events.len(),
+            self.ref_locations.len()
+        )
+    }
 }
 
 impl EventReplayer {
-    fn new() -> Self {
+    fn new_inner(main_branch_reference_name: OsString) -> Self {
         EventReplayer {
             id_counter: 0,
             events: vec![],
+            main_branch_reference_name,
             commit_history: HashMap::new(),
             ref_locations: HashMap::new(),
         }
     }
 
+    fn new(main_branch_reference_name: impl Into<OsString>) -> Self {
+        Self::new_inner(main_branch_reference_name.into())
+    }
+
     /// Construct the replayer from all the events in the database.
     ///
     /// Args:
     /// * `event_log_db`: The database to query events from.
     ///
     /// Returns: The constructed replayer.
-    pub fn from_event_log_db(event_log_db: &EventLogDb) -> anyhow::Result<Self> {
-        let mut result = EventReplayer::new();
+    pub fn from_event_log_db(
+        effects: &Effects,
+        repo: &Repo,
+        event_log_db: &EventLogDb,
+    ) -> eyre::Result<Self> {
+        let (_effects, _progress) = effects.start_operation(OperationType::ProcessEvents);
+
+        let main_branch_reference_name = repo.get_main_branch_reference()?.get_name()?;
+        let mut result = EventReplayer::new(main_branch_reference_name);
         for event in event_log_db.get_events()? {
             result.process_event(&event);
         }
@@ -677,7 +727,7 @@ impl EventReplayer {
     pub fn process_event(&mut self, event: &Event) {
         // Drop non-meaningful ref-update events.
         if let Event::RefUpdateEvent { ref_name, .. } = event {
-            if should_ignore_ref_updates(&ref_name) {
+            if should_ignore_ref_updates(ref_name) {
                 return;
             }
         }
@@ -701,22 +751,26 @@ impl EventReplayer {
                 old_commit_oid,
                 new_commit_oid,
             } => {
-                self.commit_history
-                    .entry(*old_commit_oid)
-                    .or_insert_with(Vec::new)
-                    .push(EventInfo {
-                        id,
-                        event: event.clone(),
-                        event_classification: EventClassification::Hide,
-                    });
-                self.commit_history
-                    .entry(*new_commit_oid)
-                    .or_insert_with(Vec::new)
-                    .push(EventInfo {
-                        id,
-                        event: event.clone(),
-                        event_classification: EventClassification::Show,
-                    });
+                if let MaybeZeroOid::NonZero(old_commit_oid) = old_commit_oid {
+                    self.commit_history
+                        .entry(*old_commit_oid)
+                        .or_insert_with(Vec::new)
+                        .push(EventInfo {
+                            id,
+                            event: event.clone(),
+                            event_classification: EventClassification::Hide,
+                        });
+                }
+                if let MaybeZeroOid::NonZero(new_commit_oid) = new_commit_oid {
+                    self.commit_history
+                        .entry(*new_commit_oid)
+                        .or_insert_with(Vec::new)
+                        .push(EventInfo {
+                            id,
+                            event: event.clone(),
+                            event_classification: EventClassification::Show,
+                        });
+                }
             }
 
             // A reference update doesn't indicate a change to a commit, so we
@@ -724,12 +778,12 @@ impl EventReplayer {
             // history later to find historical locations of references when
             // needed.
             Event::RefUpdateEvent {
-                ref_name, new_ref, ..
-            } => match new_ref {
-                Some(new_ref) => {
-                    self.ref_locations.insert(ref_name.clone(), new_ref.clone());
+                ref_name, new_oid, ..
+            } => match new_oid {
+                MaybeZeroOid::NonZero(new_oid) => {
+                    self.ref_locations.insert(ref_name.clone(), *new_oid);
                 }
-                None => {
+                MaybeZeroOid::Zero => {
                     self.ref_locations.remove(ref_name);
                 }
             },
@@ -789,17 +843,17 @@ impl EventReplayer {
                 timestamp,
                 event_tx_id,
                 ref_name,
-                old_ref: None,
-                new_ref: None,
+                old_oid: MaybeZeroOid::Zero,
+                new_oid: MaybeZeroOid::Zero,
                 message,
             } => {
-                let old_ref = self.ref_locations.get(&ref_name).cloned();
+                let old_oid: MaybeZeroOid = self.ref_locations.get(&ref_name).copied().into();
                 Event::RefUpdateEvent {
                     timestamp,
                     event_tx_id,
                     ref_name,
-                    old_ref,
-                    new_ref: None,
+                    old_oid,
+                    new_oid: MaybeZeroOid::Zero,
                     message,
                 }
             }
@@ -818,16 +872,16 @@ impl EventReplayer {
                     timestamp: _,
                     event_tx_id: _,
                     ref ref_name,
-                    old_ref: _,
-                    new_ref: None,
+                    old_oid: _,
+                    new_oid: MaybeZeroOid::Zero,
                     ref message,
                 },
                 Some(Event::RefUpdateEvent {
                     timestamp: _,
                     event_tx_id: _,
                     ref_name: last_ref_name,
-                    old_ref: _,
-                    new_ref: None,
+                    old_oid: _,
+                    new_oid: MaybeZeroOid::Zero,
                     message: last_message,
                 }),
             ) if ref_name == last_ref_name && message == last_message => None,
@@ -836,7 +890,7 @@ impl EventReplayer {
         }
     }
 
-    fn get_cursor_commit_history(&self, cursor: EventCursor, oid: git2::Oid) -> Vec<&EventInfo> {
+    fn get_cursor_commit_history(&self, cursor: EventCursor, oid: NonZeroOid) -> Vec<&EventInfo> {
         match self.commit_history.get(&oid) {
             None => vec![],
             Some(history) => history
@@ -857,7 +911,7 @@ impl EventReplayer {
     pub fn get_cursor_commit_visibility(
         &self,
         cursor: EventCursor,
-        oid: git2::Oid,
+        oid: NonZeroOid,
     ) -> Option<CommitVisibility> {
         let history = self.get_cursor_commit_history(cursor, oid);
         let event_info = *history.last()?;
@@ -878,7 +932,7 @@ impl EventReplayer {
     pub fn get_cursor_commit_latest_event(
         &self,
         cursor: EventCursor,
-        oid: git2::Oid,
+        oid: NonZeroOid,
     ) -> Option<&Event> {
         let history = self.get_cursor_commit_history(cursor, oid);
         let event_info = *history.last()?;
@@ -889,7 +943,7 @@ impl EventReplayer {
     ///
     /// Returns: The set of OIDs referring to commits which are thought to be
     /// active due to user action.
-    pub fn get_cursor_active_oids(&self, cursor: EventCursor) -> HashSet<git2::Oid> {
+    pub fn get_cursor_active_oids(&self, cursor: EventCursor) -> HashSet<NonZeroOid> {
         self.commit_history
             .iter()
             .filter_map(|(oid, history)| {
@@ -1014,7 +1068,7 @@ impl EventReplayer {
     ///
     /// Returns: The OID pointed to by `HEAD` at that time, or `None` if `HEAD`
     /// was never observed.
-    pub fn get_cursor_head_oid(&self, cursor: EventCursor) -> Option<git2::Oid> {
+    pub fn get_cursor_head_oid(&self, cursor: EventCursor) -> Option<NonZeroOid> {
         let cursor_event_id: usize = cursor.event_id.try_into().unwrap();
         self.events[0..cursor_event_id]
             .iter()
@@ -1023,19 +1077,9 @@ impl EventReplayer {
                 match &event {
                     Event::RefUpdateEvent {
                         ref_name,
-                        new_ref: Some(new_ref),
+                        new_oid: MaybeZeroOid::NonZero(new_oid),
                         ..
-                    } if ref_name == "HEAD" => match git2::Oid::from_str(&new_ref) {
-                        Ok(oid) => Some(oid),
-                        Err(_) => {
-                            warn!(
-                                "Expected HEAD new_ref to point to an OID; \
-                                instead pointed to: {:?}",
-                                new_ref
-                            );
-                            None
-                        }
-                    },
+                    } if ref_name == "HEAD" => Some(*new_oid),
                     Event::RefUpdateEvent { .. } => None,
 
                     // Not strictly necessary, but helps to compensate in case
@@ -1053,31 +1097,24 @@ impl EventReplayer {
     fn get_cursor_branch_oid(
         &self,
         cursor: EventCursor,
-        branch_name: &str,
-    ) -> anyhow::Result<Option<git2::Oid>> {
+        reference_name: &OsStr,
+    ) -> eyre::Result<Option<NonZeroOid>> {
         let cursor_event_id: usize = cursor.event_id.try_into().unwrap();
-        let target_ref_name = format!("refs/heads/{}", branch_name);
         let oid = self.events[0..cursor_event_id]
             .iter()
             .rev()
             .find_map(|event| match &event {
                 Event::RefUpdateEvent {
                     ref_name,
-                    new_ref: Some(new_ref),
+                    new_oid: MaybeZeroOid::NonZero(new_oid),
                     ..
-                } if *ref_name == target_ref_name => Some(new_ref),
+                } if *ref_name == reference_name => Some(*new_oid),
                 _ => None,
             });
-        match oid {
-            Some(oid) => {
-                let oid = git2::Oid::from_str(&oid)?;
-                Ok(Some(oid))
-            }
-            None => Ok(None),
-        }
+        Ok(oid)
     }
 
-    /// Get the OID of the main branch at the cursor's point in
+    /// Get the OID of the main branch at the cursor's point in time.
     ///
     /// Note that this doesn't handle the case of the user having changed their
     /// main branch configuration. That is, if it was previously `master`, and
@@ -1092,16 +1129,16 @@ impl EventReplayer {
     pub fn get_cursor_main_branch_oid(
         &self,
         cursor: EventCursor,
-        repo: &git2::Repository,
-    ) -> anyhow::Result<git2::Oid> {
-        let main_branch_name = get_main_branch_name(&repo)?;
-        let main_branch_oid = self.get_cursor_branch_oid(cursor, &main_branch_name)?;
+        repo: &Repo,
+    ) -> eyre::Result<NonZeroOid> {
+        let main_branch_reference_name = repo.get_main_branch_reference()?.get_name()?;
+        let main_branch_oid = self.get_cursor_branch_oid(cursor, &main_branch_reference_name)?;
         match main_branch_oid {
             Some(main_branch_oid) => Ok(main_branch_oid),
             None => {
                 // Assume the main branch just hasn't been observed moving yet,
                 // so its value at the current time is fine to use.
-                get_main_branch_oid(&repo)
+                repo.get_main_branch_oid()
             }
         }
     }
@@ -1119,23 +1156,21 @@ impl EventReplayer {
     pub fn get_cursor_branch_oid_to_names(
         &self,
         cursor: EventCursor,
-        repo: &git2::Repository,
-    ) -> anyhow::Result<HashMap<git2::Oid, HashSet<String>>> {
-        let mut ref_name_to_oid: HashMap<&String, git2::Oid> = HashMap::new();
+        repo: &Repo,
+    ) -> eyre::Result<HashMap<NonZeroOid, HashSet<OsString>>> {
+        let mut ref_name_to_oid: HashMap<&OsString, NonZeroOid> = HashMap::new();
         let cursor_event_id: usize = cursor.event_id.try_into().unwrap();
         for event in self.events[..cursor_event_id].iter() {
             match event {
                 Event::RefUpdateEvent {
-                    new_ref: Some(new_ref),
+                    new_oid: MaybeZeroOid::NonZero(new_oid),
                     ref_name,
                     ..
                 } => {
-                    if let Ok(oid) = git2::Oid::from_str(new_ref) {
-                        ref_name_to_oid.insert(ref_name, oid);
-                    }
+                    ref_name_to_oid.insert(ref_name, *new_oid);
                 }
                 Event::RefUpdateEvent {
-                    new_ref: None,
+                    new_oid: MaybeZeroOid::Zero,
                     ref_name,
                     ..
                 } => {
@@ -1145,25 +1180,23 @@ impl EventReplayer {
             }
         }
 
-        let mut result: HashMap<git2::Oid, HashSet<String>> = HashMap::new();
+        let mut result: HashMap<NonZeroOid, HashSet<OsString>> = HashMap::new();
         for (ref_name, ref_oid) in ref_name_to_oid.iter() {
-            match ref_name.strip_prefix("refs/heads/") {
-                None => {}
-                Some(branch_name) => {
-                    result
-                        .entry(*ref_oid)
-                        .or_insert_with(HashSet::new)
-                        .insert(String::from(branch_name));
-                }
+            if let CategorizedReferenceName::LocalBranch { .. } =
+                CategorizedReferenceName::new(ref_name)
+            {
+                result
+                    .entry(*ref_oid)
+                    .or_insert_with(HashSet::new)
+                    .insert((*ref_name).clone());
             }
         }
 
-        let main_branch_name = get_main_branch_name(&repo)?;
         let main_branch_oid = self.get_cursor_main_branch_oid(cursor, repo)?;
         result
             .entry(main_branch_oid)
             .or_insert_with(HashSet::new)
-            .insert(main_branch_name);
+            .insert(self.main_branch_reference_name.clone());
         Ok(result)
     }
 
@@ -1254,33 +1287,32 @@ mod tests {
     use super::*;
 
     use crate::testing::make_git;
-    use crate::util::get_db_conn;
     use testing::make_dummy_transaction_id;
 
     #[test]
-    fn test_drop_non_meaningful_events() -> anyhow::Result<()> {
+    fn test_drop_non_meaningful_events() -> eyre::Result<()> {
         let event_tx_id = make_dummy_transaction_id(123);
         let meaningful_event = Event::CommitEvent {
             timestamp: 0.0,
             event_tx_id,
-            commit_oid: git2::Oid::from_str("abc")?,
+            commit_oid: NonZeroOid::from_str("abc")?,
         };
-        let mut replayer = EventReplayer::new();
+        let mut replayer = EventReplayer::new("refs/heads/master");
         replayer.process_event(&meaningful_event);
         replayer.process_event(&Event::RefUpdateEvent {
             timestamp: 0.0,
             event_tx_id,
-            ref_name: String::from("ORIG_HEAD"),
-            old_ref: Some(String::from("abc")),
-            new_ref: Some(String::from("def")),
+            ref_name: OsString::from("ORIG_HEAD"),
+            old_oid: MaybeZeroOid::from_str("abc")?,
+            new_oid: MaybeZeroOid::from_str("def")?,
             message: None,
         });
         replayer.process_event(&Event::RefUpdateEvent {
             timestamp: 0.0,
             event_tx_id,
-            ref_name: String::from("CHERRY_PICK_HEAD"),
-            old_ref: None,
-            new_ref: None,
+            ref_name: OsString::from("CHERRY_PICK_HEAD"),
+            old_oid: MaybeZeroOid::Zero,
+            new_oid: MaybeZeroOid::Zero,
             message: None,
         });
 
@@ -1293,7 +1325,7 @@ mod tests {
     }
 
     #[test]
-    fn test_different_event_transaction_ids() -> anyhow::Result<()> {
+    fn test_different_event_transaction_ids() -> eyre::Result<()> {
         let git = make_git()?;
 
         git.init_repo()?;
@@ -1301,7 +1333,7 @@ mod tests {
         git.run(&["hide", "HEAD"])?;
 
         let repo = git.get_repo()?;
-        let conn = get_db_conn(&repo)?;
+        let conn = repo.get_db_conn()?;
         let event_log_db = EventLogDb::new(&conn)?;
         let events = event_log_db.get_events()?;
         let event_tx_ids: Vec<EventTransactionId> =
@@ -1339,14 +1371,14 @@ mod tests {
     }
 
     #[test]
-    fn test_advance_cursor_by_transaction() -> anyhow::Result<()> {
-        let mut event_replayer = EventReplayer::new();
+    fn test_advance_cursor_by_transaction() -> eyre::Result<()> {
+        let mut event_replayer = EventReplayer::new("refs/heads/master");
         for (timestamp, event_tx_id) in (0..).zip(&[1, 1, 2, 2, 3, 4]) {
             let timestamp: f64 = timestamp.try_into()?;
             event_replayer.process_event(&Event::UnhideEvent {
                 timestamp,
                 event_tx_id: EventTransactionId(*event_tx_id),
-                commit_oid: git2::Oid::zero(),
+                commit_oid: NonZeroOid::from_str("abc")?,
             });
         }
 
