@@ -39,6 +39,7 @@ impl<'a> ToString for Pluralize<'a> {
 }
 
 /// Glyphs to use for rendering the smartlog.
+#[derive(Clone)]
 pub struct Glyphs {
     /// Whether or not ANSI escape codes should be emitted (e.g. to render
     /// color).
@@ -86,6 +87,21 @@ pub struct Glyphs {
 
     /// Bullet-point character for a list of newline-separated items.
     pub bullet_point: &'static str,
+
+    /// Arrow character used when printing a commit cycle.
+    pub cycle_arrow: &'static str,
+
+    /// Horizontal line character used when printing a commit cycle.
+    pub cycle_horizontal_line: &'static str,
+
+    /// Vertical line character used when printing a commit cycle.
+    pub cycle_vertical_line: &'static str,
+
+    /// Corner at the upper left of the arrow used when printing a commit cycle.
+    pub cycle_upper_left_corner: &'static str,
+
+    /// Corner at the lower left of the arrow used when printing a commit cycle.
+    pub cycle_lower_left_corner: &'static str,
 }
 
 impl Glyphs {
@@ -115,11 +131,16 @@ impl Glyphs {
             commit_main_hidden: "X",
             commit_main_hidden_head: "%",
             bullet_point: "-",
+            cycle_arrow: ">",
+            cycle_horizontal_line: "-",
+            cycle_vertical_line: "|",
+            cycle_upper_left_corner: ",",
+            cycle_lower_left_corner: "`",
         }
     }
 
     /// Glyphs used for output to a TTY.
-    fn pretty() -> Self {
+    pub fn pretty() -> Self {
         Glyphs {
             should_write_ansi_escape_codes: true,
             line: "┃",
@@ -135,10 +156,25 @@ impl Glyphs {
             commit_main_hidden: "✕",
             commit_main_hidden_head: "❖",
             bullet_point: "•",
+            cycle_arrow: "ᐅ",
+            cycle_horizontal_line: "─",
+            cycle_vertical_line: "│",
+            cycle_upper_left_corner: "┌",
+            cycle_lower_left_corner: "└",
         }
     }
 }
 
+impl std::fmt::Debug for Glyphs {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(
+            f,
+            "<Glyphs pretty={:?}>",
+            self.should_write_ansi_escape_codes
+        )
+    }
+}
+
 /// Helper to build `StyledString`s by combining multiple strings (both regular
 /// `String`s and `StyledString`s).
 pub struct StyledStringBuilder {
@@ -159,24 +195,35 @@ impl StyledStringBuilder {
         }
     }
 
+    fn append_plain_inner(mut self, text: &str) -> Self {
+        self.elements.push(StyledString::plain(text));
+        self
+    }
+
     /// Append a plain-text string to the internal buffer.
-    pub fn append_plain(mut self, text: impl AsRef<str>) -> Self {
-        self.elements.push(StyledString::plain(text.as_ref()));
+    pub fn append_plain(self, text: impl AsRef<str>) -> Self {
+        self.append_plain_inner(text.as_ref())
+    }
+
+    fn append_styled_inner(mut self, text: &str, style: Style) -> Self {
+        self.elements.push(StyledString::styled(text, style));
         self
     }
 
     /// Style the provided `text` using `style`, then append it to the internal
     /// buffer.
-    pub fn append_styled(mut self, text: impl AsRef<str>, style: impl Into<Style>) -> Self {
-        self.elements
-            .push(StyledString::styled(text.as_ref(), style));
+    pub fn append_styled(self, text: impl AsRef<str>, style: impl Into<Style>) -> Self {
+        self.append_styled_inner(text.as_ref(), style.into())
+    }
+
+    fn append_inner(mut self, text: StyledString) -> Self {
+        self.elements.push(text);
         self
     }
 
     /// Directly append the provided `StyledString` to the internal buffer.
-    pub fn append(mut self, text: impl Into<StyledString>) -> Self {
-        self.elements.push(text.into());
-        self
+    pub fn append(self, text: impl Into<StyledString>) -> Self {
+        self.append_inner(text.into())
     }
 
     /// Create a new `StyledString` using all the components in the internal
@@ -191,15 +238,14 @@ impl StyledStringBuilder {
 
     /// Helper function to join a list of `StyledString`s into a single
     /// `StyledString`s, using the provided `delimiter`.
-    pub fn join(delimiter: impl Into<String>, strings: Vec<StyledString>) -> StyledString {
+    pub fn join(delimiter: &str, strings: Vec<StyledString>) -> StyledString {
         let mut result = Self::new();
         let mut is_first = true;
-        let delimiter = delimiter.into();
         for string in strings {
             if is_first {
                 is_first = false;
             } else {
-                result = result.append_plain(delimiter.clone());
+                result = result.append_plain(delimiter);
             }
             result = result.append(string);
         }
@@ -233,7 +279,7 @@ impl From<StyledStringBuilder> for StyledString {
     }
 }
 
-fn render_style_as_ansi(content: &str, style: Style) -> anyhow::Result<String> {
+fn render_style_as_ansi(content: &str, style: Style) -> eyre::Result<String> {
     let Style { effects, color } = style;
     let output = {
         use console::style;
@@ -241,10 +287,10 @@ fn render_style_as_ansi(content: &str, style: Style) -> anyhow::Result<String> {
         let output = content.to_string();
         match color.front {
             ColorType::Palette(_) => {
-                anyhow::bail!("Not implemented: using cursive palette colors")
+                eyre::bail!("Not implemented: using cursive palette colors")
             }
             ColorType::Color(Color::Rgb(..)) | ColorType::Color(Color::RgbLowRes(..)) => {
-                anyhow::bail!("Not implemented: using raw RGB colors")
+                eyre::bail!("Not implemented: using raw RGB colors")
             }
             ColorType::InheritParent | ColorType::Color(Color::TerminalDefault) => style(output),
             ColorType::Color(Color::Light(color)) => match color {
@@ -279,7 +325,7 @@ fn render_style_as_ansi(content: &str, style: Style) -> anyhow::Result<String> {
                 Effect::Reverse => output.reverse(),
                 Effect::Bold => output.bold(),
                 Effect::Italic => output.italic(),
-                Effect::Strikethrough => anyhow::bail!("Not implemented: Effect::Strikethrough"),
+                Effect::Strikethrough => eyre::bail!("Not implemented: Effect::Strikethrough"),
                 Effect::Underline => output.underlined(),
                 Effect::Blink => output.blink(),
             };
@@ -294,7 +340,7 @@ fn render_style_as_ansi(content: &str, style: Style) -> anyhow::Result<String> {
 /// style it.
 ///
 /// TODO: return something that implements `Display` instead of a `String`.
-pub fn printable_styled_string(glyphs: &Glyphs, string: StyledString) -> anyhow::Result<String> {
+pub fn printable_styled_string(glyphs: &Glyphs, string: StyledString) -> eyre::Result<String> {
     let result = string
         .spans()
         .map(|span| {
@@ -309,6 +355,6 @@ pub fn printable_styled_string(glyphs: &Glyphs, string: StyledString) -> anyhow:
                 Ok(content.to_string())
             }
         })
-        .collect::<anyhow::Result<String>>()?;
+        .collect::<eyre::Result<String>>()?;
     Ok(result)
 }
