@@ -2,45 +2,48 @@
 //!
 //! This is the basic data structure that most of branchless operates on.
 
-use std::collections::{HashMap, HashSet, VecDeque};
+use std::collections::{HashMap, HashSet};
+use std::fmt::Debug;
+use std::ops::Deref;
 
-use fn_error_context::context;
-use log::warn;
+use tracing::{instrument, warn};
 
 use crate::core::eventlog::{CommitVisibility, Event, EventCursor, EventReplayer};
 use crate::core::mergebase::MergeBaseDb;
+use crate::git::{Commit, NonZeroOid, Repo};
+use crate::tui::{Effects, OperationType};
 
 /// The OID of the repo's HEAD reference.
 #[derive(Debug)]
-pub struct HeadOid(pub Option<git2::Oid>);
+pub struct HeadOid(pub Option<NonZeroOid>);
 
 /// The OID that the repo's main branch points to.
 #[derive(Debug)]
-pub struct MainBranchOid(pub git2::Oid);
+pub struct MainBranchOid(pub NonZeroOid);
 
 /// The OIDs of any branches whose pointed-to commits should be included in the
 /// commit graph.
 #[derive(Debug)]
-pub struct BranchOids(pub HashSet<git2::Oid>);
+pub struct BranchOids(pub HashSet<NonZeroOid>);
 
 /// The OIDs of any visible commits that should be included in the commit graph.
 #[derive(Debug)]
-pub struct CommitOids(pub HashSet<git2::Oid>);
+pub struct CommitOids(pub HashSet<NonZeroOid>);
 
 /// Node contained in the smartlog commit graph.
 #[derive(Debug)]
 pub struct Node<'repo> {
     /// The underlying commit object.
-    pub commit: git2::Commit<'repo>,
+    pub commit: Commit<'repo>,
 
     /// The OID of the parent node in the smartlog commit graph.
     ///
     /// This is different from inspecting `commit.parents()`, since the smartlog
     /// will hide most nodes from the commit graph, including parent nodes.
-    pub parent: Option<git2::Oid>,
+    pub parent: Option<NonZeroOid>,
 
     /// The OIDs of the children nodes in the smartlog commit graph.
-    pub children: Vec<git2::Oid>,
+    pub children: Vec<NonZeroOid>,
 
     /// Indicates that this is a commit to the main branch.
     ///
@@ -77,68 +80,23 @@ pub struct Node<'repo> {
 }
 
 /// Graph of commits that the user is working on.
-pub type CommitGraph<'repo> = HashMap<git2::Oid, Node<'repo>>;
-
-fn find_path_to_merge_base_internal<'repo>(
-    repo: &'repo git2::Repository,
-    merge_base_db: &MergeBaseDb,
-    commit_oid: git2::Oid,
-    target_oid: git2::Oid,
-    mut visited_commit_callback: impl FnMut(git2::Oid),
-) -> anyhow::Result<Option<Vec<git2::Commit<'repo>>>> {
-    let mut queue = VecDeque::new();
-    visited_commit_callback(commit_oid);
-    queue.push_back(vec![repo.find_commit(commit_oid)?]);
-    let merge_base_oid = merge_base_db.get_merge_base_oid(repo, commit_oid, target_oid)?;
-    while let Some(path) = queue.pop_front() {
-        let last_commit = path
-            .last()
-            .expect("find_path_to_merge_base: empty path in queue");
-        if last_commit.id() == target_oid {
-            return Ok(Some(path));
-        }
-        if Some(last_commit.id()) == merge_base_oid {
-            // We've hit the common ancestor of these two commits without
-            // finding a path between them. That means it's impossible to find a
-            // path between them by traversing more ancestors. Possibly the
-            // caller passed them in in the wrong order, i.e. `commit_oid` is
-            // actually a parent of `target_oid`.
-            continue;
-        }
-
-        for parent in last_commit.parents() {
-            visited_commit_callback(parent.id());
-            let mut new_path = path.clone();
-            new_path.push(parent);
-            queue.push_back(new_path);
-        }
-    }
-    Ok(None)
+#[derive(Default)]
+pub struct CommitGraph<'repo> {
+    nodes: HashMap<NonZeroOid, Node<'repo>>,
 }
 
-/// Find a shortest path between the given commits.
-///
-/// This is particularly important for multi-parent commits (i.e. merge commits).
-/// If we don't happen to traverse the correct parent, we may end up traversing a
-/// huge amount of commit history, with a significant performance hit.
-///
-/// Args:
-/// * `repo`: The Git repository.
-/// * `commit_oid`: The OID of the commit to start at. We take parents of the
-/// provided commit until we end up at the target OID.
-/// * `target_oid`: The OID of the commit to end at.
-///
-/// Returns: A path of commits from `commit_oid` through parents to `target_oid`.
-/// The path includes `commit_oid` at the beginning and `target_oid` at the end.
-/// If there is no such path, returns `None`.
-#[context("Finding path from {:?} to {:?}", commit_oid, target_oid)]
-pub fn find_path_to_merge_base<'repo>(
-    repo: &'repo git2::Repository,
-    merge_base_db: &MergeBaseDb,
-    commit_oid: git2::Oid,
-    target_oid: git2::Oid,
-) -> anyhow::Result<Option<Vec<git2::Commit<'repo>>>> {
-    find_path_to_merge_base_internal(repo, merge_base_db, commit_oid, target_oid, |_commit| {})
+impl std::fmt::Debug for CommitGraph<'_> {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        write!(f, "<CommitGraph len={}>", self.nodes.len())
+    }
+}
+
+impl<'repo> Deref for CommitGraph<'repo> {
+    type Target = HashMap<NonZeroOid, Node<'repo>>;
+
+    fn deref(&self) -> &Self::Target {
+        &self.nodes
+    }
 }
 
 /// Find additional commits that should be displayed.
@@ -147,28 +105,35 @@ pub fn find_path_to_merge_base<'repo>(
 /// between it and the main branch, those intermediate commits should be shown
 /// (or else you won't get a good idea of the line of development that happened
 /// for this commit since the main branch).
-#[context("Walking from commits: {:?}", commit_oids)]
+#[instrument(skip(commit_oids))]
 fn walk_from_commits<'repo>(
-    repo: &'repo git2::Repository,
-    merge_base_db: &MergeBaseDb,
+    effects: &Effects,
+    repo: &'repo Repo,
+    merge_base_db: &impl MergeBaseDb,
     event_replayer: &EventReplayer,
     event_cursor: EventCursor,
     main_branch_oid: &MainBranchOid,
     commit_oids: &CommitOids,
-) -> anyhow::Result<CommitGraph<'repo>> {
-    let mut graph: CommitGraph = Default::default();
+) -> eyre::Result<CommitGraph<'repo>> {
+    let (effects, _progress) = effects.start_operation(OperationType::WalkCommits);
+
+    let mut graph: HashMap<NonZeroOid, Node> = Default::default();
 
     for commit_oid in &commit_oids.0 {
-        let commit = repo.find_commit(*commit_oid);
+        let commit = repo.find_commit(*commit_oid)?;
         let current_commit = match commit {
-            Ok(commit) => commit,
+            Some(commit) => commit,
 
             // Commit may have been garbage-collected.
-            Err(_) => continue,
+            None => continue,
         };
 
-        let merge_base_oid =
-            merge_base_db.get_merge_base_oid(repo, current_commit.id(), main_branch_oid.0)?;
+        let merge_base_oid = merge_base_db.get_merge_base_oid(
+            &effects,
+            repo,
+            current_commit.get_oid(),
+            main_branch_oid.0,
+        )?;
         let path_to_merge_base = match merge_base_oid {
             // Occasionally we may find a commit that has no merge-base with the
             // main branch. For example: a rewritten initial commit. This is
@@ -176,15 +141,18 @@ fn walk_from_commits<'repo>(
             // standalone component and hope it works out.
             None => vec![current_commit],
             Some(merge_base_oid) => {
-                let path_to_merge_base = find_path_to_merge_base(
+                let path_to_merge_base = merge_base_db.find_path_to_merge_base(
+                    &effects,
                     repo,
-                    merge_base_db,
-                    current_commit.id(),
+                    current_commit.get_oid(),
                     merge_base_oid,
                 )?;
                 match path_to_merge_base {
                     None => {
-                        warn!("No path to merge-base for commit {}", current_commit.id());
+                        warn!(
+                            current_commit_oid = ?current_commit.get_oid(),
+                            "No path to merge-base for commit",
+                        );
                         continue;
                     }
                     Some(path_to_merge_base) => path_to_merge_base,
@@ -193,29 +161,29 @@ fn walk_from_commits<'repo>(
         };
 
         for current_commit in path_to_merge_base.iter() {
-            if graph.contains_key(&current_commit.id()) {
+            if graph.contains_key(&current_commit.get_oid()) {
                 // This commit (and all of its parents!) should be in the graph
                 // already, so no need to continue this iteration.
                 break;
             }
 
             let visibility =
-                event_replayer.get_cursor_commit_visibility(event_cursor, current_commit.id());
+                event_replayer.get_cursor_commit_visibility(event_cursor, current_commit.get_oid());
             let is_visible = match visibility {
                 Some(CommitVisibility::Visible) | None => true,
                 Some(CommitVisibility::Hidden) => false,
             };
 
             let is_main = match merge_base_oid {
-                Some(merge_base_oid) => (current_commit.id() == merge_base_oid),
+                Some(merge_base_oid) => (current_commit.get_oid() == merge_base_oid),
                 None => false,
             };
 
             let event = event_replayer
-                .get_cursor_commit_latest_event(event_cursor, current_commit.id())
+                .get_cursor_commit_latest_event(event_cursor, current_commit.get_oid())
                 .cloned();
             graph.insert(
-                current_commit.id(),
+                current_commit.get_oid(),
                 Node {
                     commit: current_commit.clone(),
                     parent: None,
@@ -229,18 +197,19 @@ fn walk_from_commits<'repo>(
 
         if let Some(merge_base_oid) = merge_base_oid {
             if !graph.contains_key(&merge_base_oid) {
-                warn!("Could not find merge base OID {}", merge_base_oid);
+                warn!(?merge_base_oid, "Could not find merge base OID");
             }
         }
     }
 
     // Find immediate parent-child links.
-    let links: Vec<(git2::Oid, git2::Oid)> = graph
+    let links: Vec<(NonZeroOid, NonZeroOid)> = graph
         .iter()
         .filter(|(_child_oid, node)| !node.is_main)
         .flat_map(|(child_oid, node)| {
             node.commit
-                .parent_ids()
+                .get_parent_oids()
+                .into_iter()
                 .filter(|parent_oid| graph.contains_key(parent_oid))
                 .map(move |parent_oid| (*child_oid, parent_oid))
         })
@@ -250,27 +219,27 @@ fn walk_from_commits<'repo>(
         graph.get_mut(parent_oid).unwrap().children.push(*child_oid);
     }
 
-    Ok(graph)
+    Ok(CommitGraph { nodes: graph })
 }
 
 /// Sort children nodes of the commit graph in a standard order, for determinism
 /// in output.
 fn sort_children(graph: &mut CommitGraph) {
-    let commit_times: HashMap<git2::Oid, git2::Time> = graph
+    let commit_times: HashMap<NonZeroOid, git2::Time> = graph
         .iter()
-        .map(|(oid, node)| (*oid, node.commit.time()))
+        .map(|(oid, node)| (*oid, node.commit.get_time()))
         .collect();
-    for node in graph.values_mut() {
+    for node in graph.nodes.values_mut() {
         node.children
             .sort_by_key(|child_oid| (commit_times[child_oid], child_oid.to_string()));
     }
 }
 
 fn should_hide(
-    cache: &mut HashMap<git2::Oid, bool>,
+    cache: &mut HashMap<NonZeroOid, bool>,
     graph: &CommitGraph,
-    unhideable_oids: &HashSet<git2::Oid>,
-    oid: &git2::Oid,
+    unhideable_oids: &HashSet<NonZeroOid>,
+    oid: &NonZeroOid,
 ) -> bool {
     let result = {
         match cache.get(oid) {
@@ -319,7 +288,7 @@ fn do_remove_commits(graph: &mut CommitGraph, head_oid: &HeadOid, branch_oids: &
     }
 
     let mut cache = HashMap::new();
-    let all_oids_to_hide: HashSet<git2::Oid> = graph
+    let all_oids_to_hide: HashSet<NonZeroOid> = graph
         .keys()
         .filter(|oid| should_hide(&mut cache, graph, &unhideable_oids, oid))
         .cloned()
@@ -329,10 +298,10 @@ fn do_remove_commits(graph: &mut CommitGraph, head_oid: &HeadOid, branch_oids: &
     // appropriate.
     for oid in all_oids_to_hide {
         let parent_oid = graph[&oid].parent;
-        graph.remove(&oid);
+        graph.nodes.remove(&oid);
         match parent_oid {
             Some(parent_oid) if graph.contains_key(&parent_oid) => {
-                let children = &mut graph.get_mut(&parent_oid).unwrap().children;
+                let children = &mut graph.nodes.get_mut(&parent_oid).unwrap().children;
                 *children = children
                     .iter()
                     .filter_map(|child_oid| {
@@ -363,18 +332,21 @@ fn do_remove_commits(graph: &mut CommitGraph, head_oid: &HeadOid, branch_oids: &
 /// be set to `True` for most display-related purposes.
 ///
 /// Returns: A tuple of the head OID and the commit graph.
-#[context("Creating commit graph")]
+#[instrument]
 pub fn make_graph<'repo>(
-    repo: &'repo git2::Repository,
-    merge_base_db: &MergeBaseDb,
+    effects: &Effects,
+    repo: &'repo Repo,
+    merge_base_db: &impl MergeBaseDb,
     event_replayer: &EventReplayer,
     event_cursor: EventCursor,
     head_oid: &HeadOid,
     main_branch_oid: &MainBranchOid,
     branch_oids: &BranchOids,
     remove_commits: bool,
-) -> anyhow::Result<CommitGraph<'repo>> {
-    let mut commit_oids: HashSet<git2::Oid> = event_replayer
+) -> eyre::Result<CommitGraph<'repo>> {
+    let (effects, _progress) = effects.start_operation(OperationType::MakeGraph);
+
+    let mut commit_oids: HashSet<NonZeroOid> = event_replayer
         .get_cursor_active_oids(event_cursor)
         .into_iter()
         .collect();
@@ -384,6 +356,7 @@ pub fn make_graph<'repo>(
     }
     let commit_oids = &CommitOids(commit_oids);
     let mut graph = walk_from_commits(
+        &effects,
         repo,
         merge_base_db,
         event_replayer,
@@ -398,42 +371,35 @@ pub fn make_graph<'repo>(
     Ok(graph)
 }
 
-#[cfg(test)]
+/// The result of attempting to resolve commits.
+pub enum ResolveCommitsResult<'repo> {
+    /// All commits were successfully resolved.
+    Ok {
+        /// The commits.
+        commits: Vec<Commit<'repo>>,
+    },
 
-mod tests {
-    use super::*;
-
-    use std::collections::HashSet;
-
-    use crate::core::mergebase::MergeBaseDb;
-    use crate::testing::make_git;
-
-    #[test]
-    fn test_find_path_to_merge_base_stop_early() -> anyhow::Result<()> {
-        let git = make_git()?;
-
-        git.init_repo()?;
-        let test1_oid = git.commit_file("test1", 1)?;
-        let test2_oid = git.commit_file("test2", 2)?;
-        git.detach_head()?;
-        let test3_oid = git.commit_file("test3", 3)?;
-
-        let repo = git.get_repo()?;
-        let conn = crate::util::get_db_conn(&repo)?;
-        let merge_base_db = MergeBaseDb::new(&conn)?;
-
-        let mut seen_oids = HashSet::new();
-        let path =
-            find_path_to_merge_base_internal(&repo, &merge_base_db, test2_oid, test3_oid, |oid| {
-                seen_oids.insert(oid);
-            })?;
-        assert!(path.is_none());
-
-        println!("Seen OIDs is {:?}", &seen_oids);
-        assert!(seen_oids.contains(&test2_oid));
-        assert!(!seen_oids.contains(&test3_oid));
-        assert!(!seen_oids.contains(&test1_oid));
-
-        Ok(())
+    /// The first commit which couldn't be resolved.
+    CommitNotFound {
+        /// The identifier of the commit, as provided by the user.
+        commit: String,
+    },
+}
+
+/// Parse strings which refer to commits, such as:
+///
+/// - Full OIDs.
+/// - Short OIDs.
+/// - Reference names.
+#[instrument]
+pub fn resolve_commits(repo: &Repo, hashes: Vec<String>) -> eyre::Result<ResolveCommitsResult> {
+    let mut commits = Vec::new();
+    for hash in hashes {
+        let commit = match repo.revparse_single_commit(&hash)? {
+            Some(commit) => commit,
+            None => return Ok(ResolveCommitsResult::CommitNotFound { commit: hash }),
+        };
+        commits.push(commit)
     }
+    Ok(ResolveCommitsResult::Ok { commits })
 }
