@@ -41,5 +41,7 @@
 
 pub mod commands;
 pub mod core;
+pub mod git;
 pub mod testing;
+pub mod tui;
 pub mod util;
